from collections import defaultdict
from typing import TypeVar, Iterable, Generic

from attrs import define

from biobit import algo
from biobit.core import Interval

_T = TypeVar('_T')


@define(slots=True, frozen=True)
class Partition(Generic[_T]):
    """
    A class representing a partition of intervals.

    Attributes:
        contig: The contig of the partition.
        intervals: The intervals in the partition.
        data: The data corresponding to the intervals.
    """
    contig: str
    intervals: tuple[Interval, ...]
    data: tuple[_T, ...]

    def __attrs_post_init__(self):
        if len(self.intervals) != len(self.data):
            raise ValueError("Intervals and data lists must have the same length")

    @staticmethod
    def from_intervals(data: Iterable[_T], intervals: Iterable[Interval], maxdist: int = 1024) -> list["Partition"]:
        """
        Partition intervals into groups based on contig, strand, and their distance from each other.

        :param data: An iterable of data objects.
        :param intervals: An iterable of intervals associated with each data object.
        :param maxdist: The maximum distance between intervals for them to be considered part of the same partition.
        """
        # Group by contig
        partition = defaultdict(list)
        for d, i in zip(data, intervals):
            partition[i.contig].append((i, d))

        # Sort by start position and make final partitions
        results = []
        for contig, part in partition.items():
            part.sort(key=lambda x: x[0].rng.start)
            partitions = algo.misc.group_within(
                part,
                distance=lambda x, y: min(
                    abs(y[0].rng.start - x[0].rng.start),
                    abs(y[0].rng.end - x[0].rng.start),
                    abs(y[0].rng.start - x[0].rng.end),
                    abs(y[0].rng.end - x[0].rng.end)
                ),
                maxdist=maxdist
            )

            for p in partitions:
                pinterval, pdata = zip(*p)
                results.append(Partition(contig, pinterval, pdata))
        return results

    def __len__(self) -> int:
        return len(self.intervals)



from . import gene, transcript


class Assembly:
    def __init__(
            self,
            name: str,
            organism: str,
            version: int,
            transcripts: transcript.Descriptor | None = None,
            genes: gene.Descriptor | None = None,
    ):
        self._name = name
        self._organism = organism
        self._version = version

        self._transcripts = transcripts
        self._genes = genes

    @property
    def name(self) -> str:
        return self._name

    @property
    def organism(self) -> str:
        return self._organism

    @property
    def version(self) -> int:
        return self._version

    @property
    def transcripts(self) -> transcript.Descriptor:
        if self._transcripts is None:
            raise ValueError("Transcripts information is not available")
        return self._transcripts

    @property
    def genes(self) -> gene.Descriptor:
        if self._genes is None:
            raise ValueError("Genes information is not available")
        return self._genes



from pathlib import Path
from typing import Self

from .. import biomart
from ..abc import EnsemblAttribute


class Attribute(EnsemblAttribute):
    ID = "ID"
    Name = "name"
    Biotype = "biotype"
    Source = "source"
    Contig = "contig"
    Start = "start"
    End = "end"
    Strand = "strand"

    @property
    def colname(self) -> str:
        match self:
            case Attribute.ID:
                return "Gene stable ID version"
            case Attribute.Name:
                return "Gene name"
            case Attribute.Biotype:
                return "Gene type"
            case Attribute.Source:
                return "Source (gene)"
            case Attribute.Contig:
                return "Chromosome/scaffold name"
            case Attribute.Start:
                return "Gene start (bp)"
            case Attribute.End:
                return "Gene end (bp)"
            case Attribute.Strand:
                return "Strand"
            case _:
                raise ValueError(f"Unknown attribute {self}")

    @property
    def dtype(self):
        match self:
            case (
            Attribute.ID
            | Attribute.Name
            | Attribute.Biotype
            | Attribute.Source
            | Attribute.Contig
            ):
                return str
            case Attribute.Start | Attribute.End | Attribute.Strand:
                return int
            case _:
                raise ValueError(f"Unknown attribute {self}")

    @classmethod
    def fetch(
            cls,
            attributes: set[Self],
            organism: str,
            saveto: Path,
            force: bool,
            url: str,
            verbose: bool,
    ):
        requested = []
        for attr in attributes:
            match attr:
                case Attribute.ID:
                    requested.append("ensembl_gene_id_version")
                case Attribute.Name:
                    requested.append("external_gene_name")
                    # requested.append("external_gene_source")
                    # requested.append("external_synonym")
                case Attribute.Biotype:
                    requested.append("gene_biotype")
                case Attribute.Source:
                    requested.append("source")
                case Attribute.Contig:
                    requested.append("chromosome_name")
                case Attribute.Start:
                    requested.append("start_position")
                case Attribute.End:
                    requested.append("end_position")
                case Attribute.Strand:
                    requested.append("strand")
                case _:
                    raise ValueError(f"Unknown attribute: {attr}")

        requested = [f'<Attribute name = "{x}" />' for x in requested]
        body = "\n".join(requested)
        query = f"""
            <?xml version="1.0" encoding="UTF-8"?>
            <!DOCTYPE Query>
            <Query  virtualSchemaName = "default" formatter = "TSV" header = "1" uniqueRows = "1" count = "" datasetConfigVersion = "0.6" >
                <Dataset name = "{organism}_gene_ensembl" interface = "default" >
                    {body}
                </Dataset>
            </Query>
        """

        biomart.fetch(query, url, saveto, verbose=verbose, force=force)



import gzip
import tempfile
from pathlib import Path

from biobit.db.repmasker import RepmaskerClassification


def test_repmasker_classification():
    # Create a temporary file with some test data
    with tempfile.NamedTemporaryFile(delete=False, mode='w') as temp:
        temp.close()
        path = Path(temp.name)
        with gzip.open(path, "wt") as stream:
            stream.write("name1 class1 family1\n")
            stream.write("name2 class2 family2\n")
            stream.write("name3 class3 family3\n")

        # Initialize the RepmaskerClassification with the temporary file
        repmasker = RepmaskerClassification(path)

        # Test the classify method
        assert repmasker.classify("name1") == ("name1", "family1", "class1")
        assert repmasker.classify("name2") == ("name2", "family2", "class2")
        assert repmasker.classify("name3") == ("name3", "family3", "class3")
        assert repmasker.classify("name4") is None  # Test with a non-existent key

        # Test the names method
        assert repmasker.names() == {"name1", "name2", "name3"}

        # Test the families method
        assert repmasker.families() == {"family1", "family2", "family3"}

        # Test the classes method
        assert repmasker.classes() == {"class1", "class2", "class3"}


def test_repmasker_classification_empty():
    # Create a temporary file with no data
    with tempfile.NamedTemporaryFile(delete=False, mode='w') as temp:
        temp.close()

        # Initialize the RepmaskerClassification with the empty file
        repmasker = RepmaskerClassification(Path(temp.name))

        # Test the classify method with an empty file
        assert repmasker.classify("name1") is None

        # Test the names method with an empty file
        assert repmasker.names() == set()

        # Test the families method with an empty file
        assert repmasker.families() == set()

        # Test the classes method with an empty file
        assert repmasker.classes() == set()



from attrs import define, field

from .orientation import Orientation
from .range import Range


@define(slots=True, eq=True)
class Interval:
    contig: str
    rng: Range
    orient: Orientation = field(default=Orientation.dual)



import csv
import os
from collections import defaultdict
from io import TextIOBase
from pathlib import Path
from typing import Any, Callable

from biobit.analysis import seqproj

RowHook = Callable[[dict[str, Any]], dict[str, Any] | None]


def _reroot_fastq(row: dict[str, Any], fastq_root: os.PathLike[str] | None) -> dict[str, Any]:
    if fastq_root is None:
        return row

    fastq_root = Path(fastq_root)
    for field in ['fastq_1', 'fastq_2']:
        if field in row:
            fname = Path(row[field]).name
            row[field] = (fastq_root / fname).as_posix()
    return row


def load_bioproj(
        samplesheet: os.PathLike[str] | TextIOBase, *,
        hook: RowHook | None = None, fastq_root: os.PathLike[str] | None = None
) -> tuple[seqproj.Project, ...]:
    """
    Load a samplesheet.csv generated by nf-core/fetchngs and return a tuple of parsed bioproj.Project objects.

    :param samplesheet: The path to the samplesheet.csv file or a file-like object.
    :param hook: An optional hook function that is called for each row in the samplesheet. This might be used to filter
        out rows or modify them before parsing (e.g., correct fastq paths).
    :param fastq_root: An optional root path for fastq file names. If provided, the fastq file **names**
        in the samplesheet will be prefixed with this path before parsing. Note that only file names are used,
        the rest of the path from the samplesheet is discarded (if any).
    :return: A tuple of parsed bioproj.Project objects.
    """
    # SeqRun
    # run_accession (SRR*) -> SeqRun ind
    # instrument_model, instrument_platform -> SeqRun machine
    # library_layout -> SeqRun layout
    # read_count -> SeqRun reads
    # base_count -> SeqRun bases
    # fastq_1, fastq_2 -> SeqRun files

    # Library
    # library_source -> Library source
    # library_selection -> Library selection
    # ? -> Library stranding (always unknown)

    # Sample
    # sample_accession (SAMN*) -> Sample ind
    # scientific_name -> Sample organism
    # sample_title -> Sample tags
    # sample_description -> Sample description

    # Experiment
    # experiment_accession (SRX*) -> Experiment ind
    # experiment_title -> Experiment description

    # Project
    # study_accession (PRJNA*) -> Project ind
    # study_title -> Project description

    stream: TextIOBase
    if isinstance(samplesheet, TextIOBase):
        stream, close_later = samplesheet, False
    else:
        stream, close_later = open(samplesheet, 'r'), True

    hook = hook or (lambda x: x)

    projects: dict[str, Any] = defaultdict(lambda *_: {"samples": [], "experiments": {}, "description": None})
    for row in csv.DictReader(stream):
        # Skip empty rows
        if not row or (hooked := hook(row)) is None:
            continue

        row = _reroot_fastq(hooked, fastq_root)

        # Check for missing fields
        for field in [
            # SeqRun fields
            'run_accession', 'instrument_model', 'instrument_platform', 'library_layout', 'read_count', 'base_count',
            'fastq_1',
            # Library fields
            'library_source', 'library_selection',
            # Sample fields
            'sample_accession', 'scientific_name', 'sample_title', 'sample_description',
            # Experiment fields
            'experiment_accession', 'experiment_title',
            # Project fields
            'study_accession', 'study_title'
        ]:
            if field not in row or not row[field]:
                raise ValueError(f"Missing field {field} in the fetch-ngs samplesheet, row: {row}")
        if row['library_layout'] == 'PAIRED' and 'fastq_2' not in row:
            raise ValueError(f"Missing field fastq_2 in the fetch-ngs samplesheet for paired-end run, row: {row}")

        prjind = row['study_accession']
        project = projects[prjind]

        # Add project description
        if project["description"] is None:
            project["description"] = row['study_title']
        elif project["description"] != row['study_title']:
            raise ValueError(
                f"Project description mismatch for project {prjind}. "
                f"Expected {project['description']} but got {row['study_title']}"
            )

        # Parse attributes
        attributes = {"title": row['sample_title']}

        if row['sample_description']:
            description = row['sample_description']
        else:
            description = None

        sample = seqproj.Sample(
            ind=row['sample_accession'],
            organism=(row['scientific_name'],),
            attributes=attributes,
            description=description
        )
        project["samples"].append(sample)

        # Parse the experiment run
        library = seqproj.Library(
            source=(row['library_source'],),
            selection=(row['library_selection'],),
            stranding=seqproj.Stranding.Unknown
        )
        files: tuple[Path, ...]
        match row['library_layout']:
            case 'PAIRED':
                layout = seqproj.SeqLayout.Paired
                files = (Path(row['fastq_1']), Path(row['fastq_2']))
            case 'SINGLE':
                layout = seqproj.SeqLayout.Single
                files = (Path(row['fastq_1']),)
            case _:
                raise ValueError(
                    f"Unknown library layout {row['library_layout']} in the fetch-ngs samplesheet, row: {row}"
                )

        read_count = int(row['read_count']) if 'read_count' in row else None
        base_count = int(row['base_count']) if 'base_count' in row else None
        seqrun = seqproj.SeqRun(
            ind=row['run_accession'],
            machine=f"{row['instrument_platform']}: {row['instrument_model']}",
            layout=layout,
            files=files,
            reads=read_count,
            bases=base_count
        )

        expind = row['experiment_accession']
        if expind not in project["experiments"]:
            project["experiments"][expind] = {
                "sample": sample, "library": library, "seqruns": [seqrun], "description": row['experiment_title']
            }
        else:
            experiment = project["experiments"][expind]
            if sample != experiment["sample"]:
                raise ValueError(
                    f"Sample metadata mismatch for experiment {expind}. "
                    f"Expected {experiment['sample']} but got {sample}"
                )
            if library != experiment["library"]:
                raise ValueError(
                    f"Library metadata mismatch for experiment {expind}. "
                    f"Expected {experiment['library']} but got {library}"
                )
            if row['experiment_title'] != experiment["description"]:
                raise ValueError(
                    f"Experiment description mismatch for experiment {expind}. "
                    f"Expected {experiment['description']} but got {row['experiment_title']}"
                )
            experiment["seqruns"].append(seqrun)

    result = []
    for prjind, data in projects.items():
        experiments, samples, description = data.pop("experiments"), data.pop("samples"), data.pop("description")

        # Create bioproj.Experiments
        experiments = tuple(
            seqproj.Experiment(
                ind=expind, sample=experiment["sample"], library=experiment["library"], runs=experiment["seqruns"],
                description=experiment["description"]
            )
            for expind, experiment in experiments.items()
        )

        result.append(seqproj.Project(prjind, experiments, samples, description))

    if close_later:
        stream.close()

    return tuple(result)



from pathlib import Path

import pytest

from . import yaml
from ..experiment import Experiment
from ..library import Library, Stranding
from ..project import Project
from ..sample import Sample
from ..seqrun import SeqLayout, SeqRun


def _ensure_correctness(project: Project, serializer, deserializer):
    serialized = serializer(project)
    deserialized = deserializer(serialized)
    assert project == deserialized


@pytest.mark.parametrize("runs", [
    [SeqRun("RNA-seq", "illumina", SeqLayout.Paired, (Path("file1.fastq"), Path("file2.fastq")), 1000000, 200000000)],
    [
        SeqRun("DNA-seq", "ONT", SeqLayout.Single, (Path("file1.fastq"),)),
        SeqRun("dRNA-seq", "future-ONT", SeqLayout.Paired, (Path("file1.fastq"), Path("file2.fastq")), 1_000_000)
    ]
])
@pytest.mark.parametrize("sample", [
    Sample("S1", ("Homo sapiens", "HSV-1")),
    Sample("S2", ("Mus musculus",), {"Cells": "MEF", "Confluence": "85%"}, "My super experiment")
])
@pytest.mark.parametrize("lib", [
    Library(("transcriptome",), ("poly-A", "nuclear fraction"), Stranding.Unknown),
    Library(("DNA",), ("Total DNA",), Stranding.Reverse),
])
def test_yaml_adapter(runs, sample, lib):
    exp = Experiment("Experiment", sample, lib, runs)
    project = Project("Project", (exp,), (sample,))

    _ensure_correctness(project, yaml.dumps, yaml.loads)



import os
from io import TextIOBase

import cattrs.preconf.pyyaml

from ..experiment import Experiment
from ..library import Library
from ..project import Project
from ..sample import Sample
from ..seqrun import SeqRun

_YAML_CONVERTER = cattrs.preconf.pyyaml.make_converter()


def _unstructure_hook(exp: Experiment) -> dict:
    return {
        "ind": exp.ind,
        "sample": exp.sample.ind,
        "library": _YAML_CONVERTER.unstructure(exp.library),
        "runs": _YAML_CONVERTER.unstructure(exp.runs)
    }


def _structure_hook(data: dict, ttype: type) -> Project:
    assert ttype is Project

    samples = _YAML_CONVERTER.structure(data["samples"], tuple[Sample, ...])
    if len(samples) != len(set(s.ind for s in samples)):
        raise ValueError("Sample IDs must be unique")

    samples_mapping = {s.ind: s for s in samples}
    experiments = tuple(Experiment(
        data["ind"],
        samples_mapping[data["sample"]],
        _YAML_CONVERTER.structure(data["library"], Library),
        _YAML_CONVERTER.structure(data["runs"], tuple[SeqRun, ...])
    ) for data in data["experiments"])

    return Project(data["ind"], tuple(experiments), samples)


_YAML_CONVERTER.register_unstructure_hook(Experiment, _unstructure_hook)
_YAML_CONVERTER.register_structure_hook(Project, _structure_hook)


def load(file: os.PathLike[str] | TextIOBase) -> Project:
    if isinstance(file, TextIOBase):
        return _YAML_CONVERTER.loads(file.read(), Project)
    else:
        with open(file) as f:
            return _YAML_CONVERTER.loads(f.read(), Project)


def loads(string: str) -> Project:
    return _YAML_CONVERTER.loads(string, Project)


def dump(project: Project, saveto: os.PathLike[str] | TextIOBase) -> str:
    string = dumps(project)
    if isinstance(saveto, TextIOBase):
        saveto.write(string)
    else:
        with open(saveto, "w") as f:
            f.write(string)
    return string


def dumps(project: Project) -> str:
    return _YAML_CONVERTER.dumps(project, sort_keys=False)



from itertools import chain
from typing import Iterator

from attrs import define, field

from biobit.core import Range, Orientation
from biobit.core.strdeductor import StrandDeductor
from biobit.io import bam
from .source import Source, AlignedBlocks


@define(frozen=True, slots=True)
class SingleEndBAM(Source):
    obj: bam.Reader
    strander: StrandDeductor | None = field(default=None)

    def fetch(self, contig: str, start: int, end: int) -> Iterator[AlignedBlocks]:
        for segment in self.obj.fetch(contig, start, end):
            if self.strander is None:
                orientation = Orientation.dual
            else:
                strand = self.strander(segment)
                orientation = Orientation.from_strand(strand)

            blocks = tuple(Range(start, end) for start, end in segment.get_blocks())
            yield AlignedBlocks(blocks, orientation)

    def stats(self) -> dict:
        stats = {}
        if self.obj.statistics:
            assert self.obj.discarded is not None and self.obj.consumed is not None
            stats.update({
                "Discarded reads": self.obj.discarded.total,
                "Processed reads": self.obj.consumed.total,
            })
        return stats


@define(frozen=True, slots=True)
class PairedEndBAM(Source):
    obj: bam.PEReadsBundler
    strander: StrandDeductor | None = field(default=None)

    def fetch(self, contig: str, start: int, end: int) -> Iterator[AlignedBlocks]:
        for lmate, rmate in self.obj.fetch(contig, start, end):
            if self.strander is None:
                orientation = Orientation.dual
            else:
                strand = self.strander(lmate)
                assert strand == self.strander(rmate)
                orientation = Orientation.from_strand(strand)

            blocks = tuple(
                Range(start, end) for start, end in set(chain(lmate.get_blocks(), rmate.get_blocks()))
            )
            yield AlignedBlocks(blocks, orientation)

    def stats(self) -> dict:
        stats = {
            "Unpaired reads": self.obj.unpaired
        }

        if self.obj.reader.statistics:
            assert self.obj.reader.discarded is not None and self.obj.reader.consumed is not None
            stats.update({
                "Discarded reads": self.obj.reader.discarded.total,
                "Processed reads": self.obj.reader.consumed.total,
            })
        return stats



from pathlib import Path

import pytest

from .library import Library, Stranding
from .sample import Sample
from .seqrun import SeqLayout, SeqRun


def test_seqlayout_normalize():
    assert SeqLayout.normalize("paired") == SeqLayout.Paired
    assert SeqLayout.normalize("pe") == SeqLayout.Paired
    assert SeqLayout.normalize(SeqLayout.Paired) == SeqLayout.Paired

    assert SeqLayout.normalize("single") == SeqLayout.Single
    assert SeqLayout.normalize("se") == SeqLayout.Single
    assert SeqLayout.normalize(SeqLayout.Single) == SeqLayout.Single

    with pytest.raises(ValueError):
        SeqLayout.normalize("invalid")


def test_seqlayout_str():
    assert str(SeqLayout.Paired) == "paired-end"
    assert str(SeqLayout.Single) == "single-end"


def test_seqrun():
    run = SeqRun("run1", "illumina", "pe", ("file1.fastq", "file2.fastq"), 1000, None, "Description")
    assert run.ind == "run1"
    assert run.machine == "illumina"
    assert run.layout == SeqLayout.Paired
    assert run.files == (Path("file1.fastq"), Path("file2.fastq"))
    assert run.reads == 1000
    assert run.bases is None
    assert run.description == "Description"
    # assert repr(run) == "SeqRun(run1, illumina, paired-end, (file1.fastq, file2.fastq), 1000, None, Description)"
    # assert str(run) == ("SeqRun(run1):\n"
    #                     "\tMachine: illumina\n"
    #                     "\tLayout: paired-end\n"
    #                     "\tFiles: file1.fastq, file2.fastq\n"
    #                     "\tReads: 1000\n"
    #                     "\tBases: .\n"
    #                     "\tDescription: Description")


def test_seqrun_validators():
    ind, machine, layout, files, reads, bases, description = \
        "run1", "illumina", "pe", ("file1.fastq", "file2.fastq"), 1000, None, "Description"

    with pytest.raises(ValueError):
        SeqRun("", machine, layout, files, reads, bases, description)
    with pytest.raises(ValueError):
        SeqRun(ind, "", layout, files, reads, bases, description)
    with pytest.raises(ValueError):
        SeqRun(ind, machine, "invalid", files, reads, bases, description)
    with pytest.raises(ValueError):
        SeqRun(ind, machine, SeqLayout.Single, [], reads, bases, description)
    with pytest.raises(ValueError):
        SeqRun(ind, machine, SeqLayout.Single, ("f_1.fq", "f_2.fq"), reads, bases, description)
    with pytest.raises(ValueError):
        SeqRun(ind, machine, SeqLayout.Paired, ("f_1.fq", "f_2.fq", "f_3.fq"), reads, bases, description)
    with pytest.raises(ValueError):
        SeqRun(ind, machine, layout, files, 0, bases, description)
    with pytest.raises(ValueError):
        SeqRun(ind, machine, layout, files, None, 0, description)


def test_sample_creation():
    sample = Sample("S1", ("Homo sapiens",), {"Confluence": "75%", "Source": "HeLa"}, "Description")
    assert sample.ind == "S1"
    assert sample.organism == ("Homo sapiens",)
    assert sample.attributes == {"Confluence": "75%", "Source": "HeLa"}
    assert sample.description == "Description"
    # assert repr(sample) == f"Sample(S1, ('Homo sapiens',), {{'Confluence': '75%', 'Source': 'HeLa'}}, Description)"
    # assert str(sample) == (f"Sample(S1):\n"
    #                        f"\tOrganism: Homo sapiens\n"
    #                        f"\tAttributes: Confluence=75%, Source=HeLa\n"
    #                        f"\tDescription: Description")


def test_sample_without_description():
    sample = Sample("Mmus", ("Mus musculus",))
    assert sample.ind == "Mmus"
    assert sample.organism == ("Mus musculus",)
    assert sample.attributes == {}
    assert sample.description is None


def test_sample_with_empty_id():
    with pytest.raises(ValueError):
        Sample("", ("Homo sapiens", "HSV-1"))


def test_sample_with_empty_organism():
    with pytest.raises(ValueError):
        Sample("sample3", ())


def test_sample_with_empty_attributes():
    with pytest.raises(ValueError):
        Sample("Sample", ("Organism",), {"Confluence": ""})


def test_library_creation():
    library = Library(source=("DNA",), selection=("PCR",), stranding=Stranding.Forward)
    assert library.source == ("DNA",)
    assert library.selection == ("PCR",)
    assert library.stranding == Stranding.Forward
    # assert repr(library) == "Library(('DNA',), ('PCR',), forward)"
    # assert str(library) == ("Library:\n"
    #                         "\tSource: DNA\n"
    #                         "\tSelection: PCR\n"
    #                         "\tStranding: forward")


def test_library_stranding_normalization():
    for (strings, expected) in [
        (["f", "forward"], Stranding.Forward),
        (["r", "reverse"], Stranding.Reverse),
        (["x", "unknown"], Stranding.Unknown),
        (["u", "unstranded"], Stranding.Unstranded)
    ]:
        for s in strings:
            assert Stranding.normalize(s) == expected


def test_library_stranding_normalization_invalid():
    for string in "invalid", "":
        with pytest.raises(ValueError):
            Stranding.normalize(string)


def test_library_creation_invalid_source():
    with pytest.raises(ValueError):
        Library(source=(), selection=("PCR",), stranding=Stranding.Forward)


def test_library_creation_invalid_selection():
    with pytest.raises(ValueError):
        Library(source=("DNA",), selection=(), stranding=Stranding.Forward)



import gzip
from pathlib import Path


class RepmaskerClassification:
    def __init__(self, path: Path):
        self.path = path
        self._mapping: dict[str, tuple[str, str, str]] = {}

        with gzip.open(path, "rt") as stream:
            for line in stream:
                line = line.strip()
                if len(line) == 0:
                    continue
                name, cls, family = line.split()
                assert name not in self._mapping, name
                self._mapping[name] = (name, family, cls)

    def classify(self, repname: str) -> tuple[str, str, str] | None:
        return self._mapping.get(repname, None)

    def names(self) -> set[str]:
        return set(x[0] for x in self._mapping.values())

    def families(self) -> set[str]:
        return set(x[1] for x in self._mapping.values())

    def classes(self) -> set[str]:
        return set(x[2] for x in self._mapping.values())



from .joblib_counter import JoblibMultiReadsCounter
from .partition import Partition
from .reads_counter import MultiReadsCounter

__all__ = ["Partition", "MultiReadsCounter", "JoblibMultiReadsCounter"]



from enum import Enum
from typing import Literal, TypeVar, Generic

from attrs import define, field


class Strand(Enum):
    fwd = "+"
    rev = "-"

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value

    @classmethod
    def normalize(cls, strand: 'str | int | Strand') -> 'Strand':
        match strand:
            case "+" | 1:
                return Strand.fwd
            case "-" | -1:
                return Strand.rev
            case _ if isinstance(strand, cls):
                return strand
            case _:
                raise ValueError(f"Unknown strand: {strand}")


StrandLike = Strand | Literal["+", "-", 1, -1]

_T = TypeVar("_T")


@define(slots=True, frozen=True, eq=True, repr=True, hash=True)
class Stranded(Generic[_T]):
    fwd: _T = field()
    rev: _T = field()

    def with_rev(self, value: _T) -> 'Stranded[_T]':
        return Stranded(self.fwd, value)

    def with_fwd(self, value: _T) -> 'Stranded[_T]':
        return Stranded(value, self.rev)

    def __iter__(self):
        yield from (self.fwd, self.rev)



from attrs import define, field

from .library import Library
from .sample import Sample
from .seqrun import SeqRun


@define(hash=True, slots=True, frozen=True, eq=True, order=True, repr=True, str=True)
class Experiment:
    """
    A class representing a biological experiment, which consists of a biological material isolated from a sample,
    turned into a sequencing library, and sequenced one or more times.

    Attributes
    ----------
    ind : str
        Experiment ID, must be unique within the project.
    sample : Sample
        The biological sample from which the library was generated.
    library : Library
        The library generated from the sample.
    runs : tuple[SeqRun, ...]
        The sequencing runs performed on the library.
    description : str
        A description of the experiment, if available.
    """
    ind: str = field()
    sample: Sample = field()
    library: Library = field()
    runs: tuple[SeqRun, ...] = field(converter=lambda x: tuple(x))
    # Optional fields
    description: str | None = field(default=None)

    @ind.validator
    def check_ind(self, _, value):
        if not value:
            raise ValueError("Experiment ID must be specified")

    @runs.validator
    def check_runs(self, _, value):
        if not value:
            raise ValueError("At least one sequencing run must be specified")

        # Ensure that run IDs are unique within the experiment
        cnts = {}
        for x in value:
            cnts[x.ind] = cnts.get(x.ind, 0) + 1
        nonunique = {k for k, v in cnts.items() if v > 1}
        if nonunique:
            raise ValueError(f"Run IDs must be unique within the experiment, found duplicates: {nonunique}")

    @description.validator
    def check_description(self, _, value):
        if value is not None and not value:
            raise ValueError("If specified, description must be non-empty. Use None to indicate lack of description")



from . import source, resolve
from .counter import MultiReadsCounter
from .resolve import Resolution
from .source import Source

__all__ = ['source', 'resolve', 'counter', 'Source', 'Resolution', 'MultiReadsCounter']



import pytest

from biobit.core import Range
from biobit.core.range import OverlapInfo


def test_add_some():
    range1 = Range(1, 5)
    result = range1 + 2
    assert result.start == 3
    assert result.end == 7


def test_add_zero():
    range1 = Range(1, 5)
    result = range1 + 0
    assert result.start == 1
    assert result.end == 5


def test_sub():
    range1 = Range(1, 5)
    result = range1 - 1
    assert result.start == 0
    assert result.end == 4


def test_len():
    range1 = Range(1, 5)
    assert len(range1) == 4


def test_eq_true():
    range1 = Range(1, 5)
    assert range1 == Range(1, 5)


def test_eq_false():
    range1 = Range(1, 5)
    range2 = Range(3, 7)
    assert range1 != range2


def test_contains_true():
    range1 = Range(1, 5)
    range2 = Range(2, 4)
    assert range2 in range1


def test_contains_false():
    range1 = Range(1, 5)
    range2 = Range(5, 7)
    assert range2 not in range1


def test_overlap_some():
    range1 = Range(1, 5)
    range2 = Range(3, 7)
    overlap = range1.overlap(range2)
    assert isinstance(overlap, OverlapInfo)
    assert overlap.size == 2
    assert overlap.rng.start == 3
    assert overlap.rng.end == 5


def test_overlap_none():
    range1 = Range(1, 5)
    range2 = Range(6, 10)
    overlap = range1.overlap(range2)
    assert overlap is None


def test_overlap_none_touching():
    range1 = Range(1, 5)
    range2 = Range(5, 10)
    overlap = range1.overlap(range2)
    assert overlap is None


def test_overlap_with_start_end():
    range1 = Range(1, 5)
    overlap = range1.overlap(start=3, end=7)
    assert isinstance(overlap, OverlapInfo)
    assert overlap.size == 2
    assert overlap.rng.start == 3
    assert overlap.rng.end == 5


def test_overlap_without_other_or_start_end():
    range1 = Range(1, 5)
    with pytest.raises(ValueError):
        range1.overlap()


def test_add_with_non_int():
    range1 = Range(1, 5)
    with pytest.raises(TypeError):
        range1 + 'a'


def test_init_with_start_greater_than_end():
    with pytest.raises(ValueError):
        Range(5, 1)



from dataclasses import dataclass
from pathlib import Path
from typing import Literal

import pandas as pd

from .attributes import Attribute
from ..abc import EnsemblDescriptor


@dataclass(frozen=True, slots=True)
class Record:
    id: str
    # Attributes
    name: str | None = None
    biotype: str | None = None
    source: str | None = None
    # Coordinates
    contig: str | None = None
    start: int | None = None
    end: int | None = None
    strand: Literal["+", "-"] | None = None

    @property
    def coordinates(self) -> str | None:
        if (
                self.contig is None
                or self.start is None
                or self.end is None
                or self.strand is None
        ):
            return None
        return f"{self.contig}:{self.start}-{self.end}({self.strand})"


class Descriptor(EnsemblDescriptor):
    def __init__(self, attributes: set[Attribute], cache: Path):
        self.attributes = attributes
        self.cache = cache
        self._records: dict[str, Record] = {}

        if attributes and Attribute.ID not in attributes:
            raise ValueError(
                f"Attribute {Attribute.ID} must be included when querying transcript attributes"
            )

        dtypes = {x.colname: x.dtype for x in attributes}
        df = pd.read_csv(
            self.cache, sep="\t", dtype=dtypes, usecols=list(dtypes.keys())
        )

        if Attribute.Strand in attributes:
            assert set(df[Attribute.Strand.colname].unique()) == {1, -1}, df[
                Attribute.Strand.colname
            ].unique()
            df[Attribute.Strand.colname] = df[Attribute.Strand.colname].apply(
                lambda x: "+" if x > 0 else "-"
            )

        # Create records
        records = [Record(id=id) for id in df[Attribute.ID.colname]]
        mapping = {
            Attribute.Name: "name",
            Attribute.Biotype: "biotype",
            Attribute.Source: "source",
            Attribute.Contig: "contig",
            Attribute.Start: "start",
            Attribute.End: "end",
            Attribute.Strand: "strand",
        }
        for attr in attributes - {Attribute.ID}:
            for r, val in zip(records, df[attr.colname]):
                val = val if not pd.isna(val) else None
                object.__setattr__(r, mapping[attr], val)

        for r in records:
            self._records[r.id] = r
            self._records[r.id.split(".")[0]] = r

    def record(self, gid: str) -> Record | None:
        return self._records.get(gid, None)

    # # Gene id -> gene coordinates
    # self._id2coords = None
    # if all(attr in attributes for attr in [Attribute.Contig, Attribute.Start, Attribute.End, Attribute.Strand]):
    #     self._id2coords = {}
    #     cols = [Attribute.Contig.colname, Attribute.Start.colname, Attribute.End.colname, Attribute.Strand.colname]
    #     for id in idcols:
    #         self._id2coords |= dict(zip(df[id], df[cols].itertuples(index=False, name=None)))
    #
    # # Gene id -> gene biotype
    # self._id2biotype = None
    # if Attribute.Biotype in attributes:
    #     self._id2biotype = {}
    #     for id in idcols:
    #         self._id2biotype |= dict(zip(df[id], df[Attribute.Biotype.colname]))
    #
    # # Gene id -> gene name; Gene name -> gene id
    # self._id2name, self._name2id = None, None
    # if Attribute.Name in attributes:
    #     self._id2name = {}
    #     for id in idcols:
    #         self._id2name |= dict(zip(df[id], df[Attribute.Name.colname]))
    #     self._name2id = dict(zip(df[Attribute.Name.colname], df[Attribute.ID.colname]))

    # Synonyms -> gene id
    #     self._synonym2id = self._name2id.copy()
    #     self._synonym2id.update(
    #         {k: v for k, v in zip(df['Gene Synonym'].values, df['Gene stable ID version'].values) if not pd.isna(k)}
    #     )
    #     # self._synonym2id.update({
    #     #     k: v
    #     #     for k, v in zip(df['NCBI gene (formerly Entrezgene) accession'].values, df['Gene stable ID version'].values)
    #     # })
    #     self._synonym2id.update({k.lower(): v for k, v in self._synonym2id.items()})
    #     self._synonym2id.update({k.upper(): v for k, v in self._synonym2id.items()})
    #
    # def id2coords(self, id: str) -> Optional[Tuple[str, int, int, str]]:
    #     return self._getter(self._id2coords, id)
    #
    # def id2biotype(self, id: str) -> Optional[str]:
    #     return self._getter(self._id2biotype, id)
    #
    # def id2name(self, id: str) -> Optional[str]:
    #     return self._getter(self._id2name, id)
    #
    # def id2transcripts(self, id: str) -> Optional[Tuple[str, ...]]:
    #     return self._getter(self._id2transcripts, id)
    #
    # def name2id(self, name: str) -> Optional[str]:
    #     return self._getter(self._name2id, name)
    #
    # def synonym2id(self, synonym: str) -> Optional[str]:
    #     gid = self._getter(self._synonym2id, synonym)
    #     if gid is None:
    #         gid = self._getter(self._synonym2id, synonym.lower())
    #     if gid is None:
    #         gid = self._getter(self._synonym2id, synonym.upper())
    #     return gid



from collections import defaultdict
from collections.abc import Callable

from attrs import define, field

from .experiment import Experiment
from .sample import Sample
from .seqrun import SeqRun


@define(hash=True, slots=True, frozen=True, eq=True, order=True, repr=True, str=True)
class Project:
    """
    A class representing a biological project, which consists of one or more biological experiments.

    Attributes
    ----------
    ind : str
        Project ID, must be globally unique.
    experiments : tuple[Experiment, ...]
        The experiments in the project.
    samples : tuple[Sample, ...]
        The samples in the project.
    description : str
        A description of the project, if available.
    """
    ind: str = field()
    experiments: tuple[Experiment, ...] = field(converter=lambda x: tuple(x))
    samples: tuple[Sample, ...] = field(converter=lambda x: tuple(x))
    # Optional fields
    description: str | None = field(default=None)

    @ind.validator
    def check_ind(self, _, value):
        if not value:
            raise ValueError("Project ID must be specified")

    @experiments.validator
    def check_experiments(self, _, value):
        if not value:
            raise ValueError("At least one experiment must be specified")

    @samples.validator
    def check_samples(self, _, value):
        if not value:
            raise ValueError("At least one sample must be specified")

    @description.validator
    def check_description(self, _, value):
        if value is not None and not value:
            raise ValueError("If specified, description must be non-empty. Use None to indicate lack of description")

    def __attrs_post_init__(self):
        # Ensure that experiment IDs are unique within the project
        cnts = {}
        for x in self.experiments:
            cnts[x.ind] = cnts.get(x.ind, 0) + 1
        nonunique = {k for k, v in cnts.items() if v > 1}
        if nonunique:
            raise ValueError(f"Experiment IDs must be unique within the project, found duplicates: {nonunique}")

        # Ensure that run IDs are unique within the project
        cnts = {}
        for x in self.experiments:
            for y in x.runs:
                cnts[y.ind] = cnts.get(y.ind, 0) + 1
        nonunique = {k for k, v in cnts.items() if v > 1}
        if nonunique:
            raise ValueError(f"Run IDs must be unique within the project, found duplicates: {nonunique}")

        # Gather all samples from the experiments
        samples = defaultdict(list)
        for x in self.experiments:
            samples[x.sample.ind].append(x.sample)

        # Ensure that samples with the same ID are identical
        for k, v in samples.items():
            if len(v) > 1 and any(x != v[0] for x in v[1:]):
                raise ValueError(f"Samples with the same ID must be identical, found differences for sample {k}: {v}")

    def subsample(
            self,
            newind: str | None = None,
            experiments: Callable[[Experiment], bool] | None = None,
            samples: Callable[[Experiment, Sample], bool] | None = None,
            runs: Callable[[Experiment, SeqRun], bool] | None = None
    ) -> "Project":
        """
        Subsample the project based on the specified criteria.
        :param newind: new project ID. If None, the original ID is used.
        :param experiments: callable that takes an experiment and returns True if it should be included, False
            otherwise. If None, all experiments are included.
        :param samples: callable that takes an experiment and a sample and returns True if the sample should be
            included, False otherwise. If None, all samples are included.
        :param runs: callable that takes an experiment and a sequencing run and returns True if the run should be
            included, False otherwise. If None, all runs are included.
        :return: a new Project instance including only the specified experiments, samples, and runs
        """
        raise ValueError("Not implemented")



from typing import Callable, Literal

from pysam import AlignedSegment

from .strand import Strand
from biobit.analysis import seqproj

StrandDeductor = Callable[[AlignedSegment], Strand]


def get(protocol: Literal["f/s", "s/f", "f", "s"]) -> StrandDeductor:
    match protocol:
        case "f/s":
            return _fs
        case "s/f":
            return _sf
        case "f":
            return _f
        case "s":
            return _s
        case _:
            raise ValueError(f"Unknown stranding protocol: {protocol}")


def from_bioproj(experiment: seqproj.Experiment) -> StrandDeductor:
    runs = list(experiment.runs)
    layout = runs[0].layout
    for r in runs[1:]:
        if r.layout != layout:
            raise ValueError(
                f"All runs must have the same layout. Found different layouts: {layout} and {r.layout} in {experiment}"
            )

    match (layout, experiment.library.stranding):
        case (seqproj.SeqLayout.Single, seqproj.Stranding.Forward):
            return _s
        case (seqproj.SeqLayout.Single, seqproj.Stranding.Reverse):
            return _f
        case (seqproj.SeqLayout.Paired, seqproj.Stranding.Forward):
            return _fs
        case (seqproj.SeqLayout.Paired, seqproj.Stranding.Reverse):
            return _sf
        case _:
            raise ValueError(
                f"Unsupported layout/stranding combination: {layout}/{experiment.library.stranding}\n"
                f"bioproj.Experiment: {experiment}"
            )


def _f(read: AlignedSegment) -> Strand:
    # Flip the mapped strand
    match read.is_reverse:
        case True:
            return Strand.fwd
        case False:
            return Strand.rev
        case _:
            assert False, "Unreachable"


def _s(read: AlignedSegment) -> Strand:
    # Keep the mapped strand
    match read.is_reverse:
        case True:
            return Strand.rev
        case False:
            return Strand.fwd
        case _:
            assert False, "Unreachable"


def _fs(read: AlignedSegment) -> Strand:
    match (read.is_read1, read.is_reverse):
        # First mate -> flip its strand
        case (True, True):
            return Strand.fwd
        case (True, False):
            return Strand.rev
        # Second mate -> keep its strand
        case (False, True):
            return Strand.rev
        case (False, False):
            return Strand.fwd
        case _:
            assert False, "Unreachable"


def _sf(read: AlignedSegment) -> Strand:
    match (read.is_read1, read.is_reverse):
        # First mate -> keep its strand
        case (True, True):
            return Strand.rev
        case (True, False):
            return Strand.fwd
        # Second mate -> flip its strand
        case (False, True):
            return Strand.fwd
        case (False, False):
            return Strand.rev
        case _:
            assert False, "Unreachable"



from collections.abc import Iterable
from pathlib import Path
from typing import TypeVar, Callable

from biobit import io
from biobit.analysis import seqproj
from biobit.core import strdeductor
from .bam import SingleEndBAM, PairedEndBAM
from .source import Source
from ...nfcore import rnaseq

_K = TypeVar('_K')


def default_se_bam_reader(path: Path) -> io.bam.Reader:
    return io.bam.Reader(path, inflags=0, exflags=2564, minmapq=0, statistics=True)


def default_pe_bam_reader(path: Path) -> io.bam.PEReadsBundler:
    reader = io.bam.Reader(path, inflags=3, exflags=2564, minmapq=0, statistics=True)
    return io.bam.PEReadsBundler(reader)


def default_key(prj: rnaseq.Project, experiment: rnaseq.Experiment) -> tuple[str, str]:
    return prj.ind, experiment.ind


def from_rnaseq(
        projects: Iterable[rnaseq.Project], *,
        se_bam: Callable[[Path], io.bam.Reader] = default_se_bam_reader,
        pe_bam: Callable[[Path], io.bam.PEReadsBundler] = default_pe_bam_reader,
        key: Callable[[rnaseq.Project, rnaseq.Experiment], _K] = default_key,  # type: ignore
) -> dict[_K, Source]:
    results: dict[_K, Source] = {}
    for prj in projects:
        for experiment in prj.experiments:
            stranding = strdeductor.from_bioproj(experiment.source)

            runs = list(experiment.runs)
            layout = runs[0].layout
            for r in runs[1:]:
                if r.layout != layout:
                    raise ValueError(
                        f"All runs must have the same layout. Found different layouts: {layout} and {r.layout} in {experiment}"
                    )

            record = key(prj, experiment)
            if record in results:
                raise ValueError(f"Duplicate key: {record}")

            match layout:
                case seqproj.SeqLayout.Single:
                    results[record] = SingleEndBAM(se_bam(experiment.bam), stranding)
                case seqproj.SeqLayout.Paired:
                    results[record] = PairedEndBAM(pe_bam(experiment.bam), stranding)
                case _:
                    raise ValueError(f"Unsupported layout: {layout}")
    return results



import os
from io import TextIOBase
from typing import Callable

from biobit.analysis import seqproj
from . import descriptor

__all__ = ["from_bioproj"]


def from_bioproj(
        project: seqproj.Project, saveto: os.PathLike[str] | TextIOBase, *,
        bioexp2desc: Callable[[seqproj.Experiment], str] = descriptor.from_bioexp
) -> str:
    """
    Converts a given bio project into an input file (design) for the nf-core/rnaseq pipeline.

    The function generates a CSV table with columns: "sample", "fastq_1", "fastq_2", "strandedness".
    Each row corresponds to a run in the project's experiments. The "sample" column uses the experiment ID as the sample name.
    The "strandedness" column is derived from the experiment's library stranding, with "auto" used for unknown stranding.
    The "fastq_1" and "fastq_2" columns are populated with the file paths from the run's files.
    For single-end runs, "fastq_2" is left empty.

    :param project: The project object to be converted.
    :param saveto: The destination for the output. This can be a file path (as a string or Path object) or a TextIO stream.
    :param bioexp2desc: A function that converts a bioproj experiment into a human-readable descriptor (sample column).
    :return: The content of the generated input file.
    """
    # Columns
    lines = ["sample,fastq_1,fastq_2,strandedness"]
    for exp in project.experiments:
        descriptor = bioexp2desc(exp)

        # Convert stranding to the nf-core/rnaseq format
        match exp.library.stranding:
            case seqproj.Stranding.Unknown:
                stranding = "auto"
            case other:
                stranding = other.value

        for run in exp.runs:
            match run.layout:
                case seqproj.SeqLayout.Paired:
                    fastq1, fastq2 = run.files[0].as_posix(), run.files[1].as_posix()
                case seqproj.SeqLayout.Single:
                    fastq1, fastq2 = run.files[0].as_posix(), ""
                case _:
                    raise ValueError(f"Unsupported sequencing layout: {run.layout}")

            lines.append(f"{descriptor},{fastq1},{fastq2},{stranding}")

    content = "\n".join(lines)

    # Save & return the content
    if isinstance(saveto, TextIOBase):
        saveto.write(content)
    else:
        with open(saveto, "w") as stream:
            stream.write(content)
    return content



from collections.abc import Iterable
from typing import Callable

from attrs import define, field

from .experiment import Experiment
from .library import Library
from .project import Project
from .sample import Sample
from .seqrun import SeqRun, SeqLayout


@define(slots=True)
class FilteringSchema:
    _project_filters: list[Callable[[Project], bool]] = field(factory=list, init=False)
    _experiment_filters: list[Callable[[Project, Experiment], bool]] = field(factory=list, init=False)
    _sample_filters: list[Callable[[Project, Sample], bool]] = field(factory=list, init=False)
    _run_filters: list[Callable[[Project, Experiment, SeqRun], bool]] = field(factory=list, init=False)
    _library_filters: list[Callable[[Project, Experiment, Library], bool]] = field(factory=list, init=False)

    def with_project_filters(self, *fns: Callable[[Project], bool]) -> "FilteringSchema":
        self._project_filters.extend(fns)
        return self

    def with_experiment_filters(self, *fns: Callable[[Project, Experiment], bool]) -> "FilteringSchema":
        self._experiment_filters.extend(fns)
        return self

    def with_sample_filters(self, *fns: Callable[[Project, Sample], bool]) -> "FilteringSchema":
        self._sample_filters.extend(fns)
        return self

    def with_run_filters(self, *fns: Callable[[Project, Experiment, SeqRun], bool]) -> "FilteringSchema":
        self._run_filters.extend(fns)
        return self

    def with_library_filters(self, *fns: Callable[[Project, Experiment, Library], bool]) -> "FilteringSchema":
        self._library_filters.extend(fns)
        return self

    def with_project_ind(self, allowed: Iterable[str] | str) -> "FilteringSchema":
        if isinstance(allowed, str):
            options = {allowed}
        else:
            options = set(allowed)
        return self.with_project_filters(lambda prj: prj.ind in options)

    def with_attribute(self, attribute: str, allowed: Iterable[str] | str) -> "FilteringSchema":
        if isinstance(allowed, str):
            options = {allowed}
        else:
            options = set(allowed)
        return self.with_sample_filters(lambda _, sample: sample.attributes.get(attribute) in options)

    def with_seqlayout(self, allowed: Iterable[SeqLayout] | SeqLayout) -> "FilteringSchema":
        if isinstance(allowed, SeqLayout):
            options = {SeqLayout(allowed)}
        else:
            options = set(allowed)
        return self.with_run_filters(lambda _, __, run: run.layout in options)

    def __call__(self, projects: Iterable[Project]) -> Iterable[Project]:
        for prj in projects:
            if not all(fn(prj) for fn in self._project_filters):
                continue

            # Select only samples that passed the filters
            samples = {sample.ind for sample in prj.samples if all(fn(prj, sample) for fn in self._sample_filters)}
            if not samples:
                continue

            # Select only experiments that passed the filters
            passed_experiments = []
            for exp in prj.experiments:
                # Check if the sample passed the filters
                if exp.sample.ind not in samples:
                    continue
                # Check if library pass the filters
                if not all(fn(prj, exp, exp.library) for fn in self._library_filters):
                    continue
                # Check if at least one run pass the filters
                runs = tuple(run for run in exp.runs if all(fn(prj, exp, run) for fn in self._run_filters))
                if not runs:
                    continue
                # Add the experiment to the list of passed experiments
                passed_experiments.append(Experiment(exp.ind, exp.sample, exp.library, runs, exp.description))

            # Keep only samples that are present in at least one experiment
            samples = {exp.sample.ind for exp in passed_experiments}

            if passed_experiments:
                yield Project(
                    prj.ind,
                    experiments=tuple(passed_experiments),
                    samples=tuple(sample for sample in prj.samples if sample.ind in samples),
                    description=prj.description
                )



from pathlib import Path
from typing import Any, Self

from biobit.db.ensembl.abc import EnsemblAttribute
from .. import biomart


class Attribute(EnsemblAttribute):
    ID = "ID"
    Gene = "gene"
    Name = "name"
    Biotype = "biotype"
    Contig = "contig"
    Start = "start"
    End = "end"
    Strand = "strand"
    Source = "source"
    SupportLevel = "support level"
    MANESelect = "MANE Select"
    MANEPlusClinical = "MANE Plus Clinical"
    GENCODEbasic = "GENCODE basic"
    EnsemblCanonical = "Ensembl Canonical"

    @property
    def colname(self) -> str:
        match self:
            case Attribute.ID:
                return "Transcript stable ID version"
            case Attribute.Gene:
                return "Gene stable ID version"
            case Attribute.Name:
                return "Transcript name"
            case Attribute.Biotype:
                return "Transcript type"
            case Attribute.SupportLevel:
                return "Transcript support level (TSL)"
            case Attribute.Contig:
                return "Chromosome/scaffold name"
            case Attribute.Start:
                return "Transcript start (bp)"
            case Attribute.End:
                return "Transcript end (bp)"
            case Attribute.Strand:
                return "Strand"
            case Attribute.MANESelect:
                return "RefSeq match transcript (MANE Select)"
            case Attribute.MANEPlusClinical:
                return "RefSeq match transcript (MANE Plus Clinical)"
            case Attribute.GENCODEbasic:
                return "GENCODE basic annotation"
            case Attribute.EnsemblCanonical:
                return "Ensembl Canonical"
            case Attribute.Source:
                return "Source (transcript)"
            case _:
                raise ValueError(f"Unknown attribute {self}")

    @property
    def dtype(self) -> Any:
        match self:
            case (
            Attribute.ID
            | Attribute.Gene
            | Attribute.Name
            | Attribute.Biotype
            | Attribute.Source
            | Attribute.SupportLevel
            ):
                return str
            case (
            Attribute.MANESelect
            | Attribute.MANEPlusClinical
            | Attribute.GENCODEbasic
            ):
                return str
            case Attribute.EnsemblCanonical:
                return "boolean"
            case Attribute.Contig:
                return str
            case Attribute.Start | Attribute.End | Attribute.Strand:
                return int
            case _:
                raise ValueError(f"Unknown attribute {self}")

    @classmethod
    def fetch(
            cls,
            attributes: set[Self],
            organism: str,
            saveto: Path,
            force: bool,
            url: str,
            verbose: bool,
    ):
        requested = []
        for attr in attributes:
            match attr:
                case Attribute.ID:
                    requested.append("ensembl_transcript_id_version")
                case Attribute.Gene:
                    requested.append("ensembl_gene_id_version")
                case Attribute.Name:
                    requested.append("external_transcript_name")
                case Attribute.Biotype:
                    requested.append("transcript_biotype")
                case Attribute.SupportLevel:
                    requested.append("transcript_tsl")
                case Attribute.Contig:
                    requested.append("chromosome_name")
                case Attribute.Start:
                    requested.append("transcript_start")
                case Attribute.End:
                    requested.append("transcript_end")
                case Attribute.Strand:
                    requested.append("strand")
                case Attribute.MANESelect:
                    requested.append("transcript_mane_select")
                case Attribute.MANEPlusClinical:
                    requested.append("transcript_mane_plus_clinical")
                case Attribute.GENCODEbasic:
                    requested.append("transcript_gencode_basic")
                case Attribute.EnsemblCanonical:
                    requested.append("transcript_is_canonical")
                case Attribute.Source:
                    requested.append("transcript_source")
                case _:
                    raise ValueError(f"Unknown attribute {attr}")

        requested = [f'<Attribute name = "{x}" />' for x in requested]
        body = "\n".join(requested)
        query = f"""
            <?xml version="1.0" encoding="UTF-8"?>
            <!DOCTYPE Query>
            <Query  virtualSchemaName = "default" formatter = "TSV" header = "1" uniqueRows = "1" count = "" datasetConfigVersion = "0.6" >
                <Dataset name = "{organism}_gene_ensembl" interface = "default" >
                    {body}
                </Dataset>
            </Query>
        """

        biomart.fetch(query, url, saveto, verbose=verbose, force=force)



import pytest

from biobit.ds.gindex.overlap import Overlap, OverlapSteps
from biobit.core import Range


@pytest.fixture
def overlap_instance():
    rng = Range(0, 10)
    intervals = [Range(1, 3), Range(4, 6), Range(7, 9)]
    annotations = ['a', 'b', 'c']
    return Overlap(rng, intervals, annotations)


@pytest.fixture
def overlap_steps_instance():
    rng = Range(0, 10)
    boundaries = [Range(0, 2), Range(2, 4), Range(4, 6), Range(6, 8), Range(8, 10)]
    annotations = [{'a'}, {'b'}, {'c'}, {'d'}, {'e'}]
    return OverlapSteps(rng, boundaries, annotations)


def test_overlap_len(overlap_instance):
    assert len(overlap_instance) == 3


def test_overlap_iter(overlap_instance):
    expected = [(Range(1, 3), 'a'), (Range(4, 6), 'b'), (Range(7, 9), 'c')]
    assert list(overlap_instance) == expected


def test_overlap_to_steps(overlap_instance):
    steps = overlap_instance.to_steps()
    assert isinstance(steps, OverlapSteps)
    assert steps.rng == overlap_instance.rng
    assert steps.boundaries == [
        Range(0, 1), Range(1, 3), Range(3, 4), Range(4, 6), Range(6, 7), Range(7, 9), Range(9, 10)
    ]
    assert steps.annotations == [set(), {'a'}, set(), {'b'}, set(), {'c'}, set()]


def test_overlap_steps_len(overlap_steps_instance):
    assert len(overlap_steps_instance) == 5


def test_overlap_steps_iter(overlap_steps_instance):
    expected = [
        (Range(0, 2), {'a'}), (Range(2, 4), {'b'}), (Range(4, 6), {'c'}), (Range(6, 8), {'d'}),
        (Range(8, 10), {'e'})
    ]
    assert list(overlap_steps_instance) == expected


def test_overlap_to_steps_nested_intervals(overlap_instance):
    # Create an Overlap instance with nested intervals
    rng = Range(0, 10)
    intervals = [Range(1, 9), Range(2, 8), Range(3, 7)]
    annotations = ['a', 'b', 'c']
    overlap = Overlap(rng, intervals, annotations)

    # Call the to_steps method
    steps = overlap.to_steps()

    # Assert that the returned OverlapSteps instance has the correct rng, boundaries, and annotations
    assert isinstance(steps, OverlapSteps)
    assert steps.rng == overlap.rng
    assert steps.boundaries == [
        Range(0, 1), Range(1, 2), Range(2, 3), Range(3, 7), Range(7, 8), Range(8, 9), Range(9, 10)
    ]
    assert steps.annotations == [set(), {'a'}, {'a', 'b'}, {'a', 'b', 'c'}, {'a', 'b'}, {'a'}, set()]


def test_overlap_empty_intervals(overlap_instance):
    # Create an Overlap instance with no intervals
    rng = Range(0, 10)
    overlap = Overlap(rng, [], [])

    # Assert that the length is 0
    assert len(overlap) == 0

    # Assert that converting to steps results in a single boundary with no annotations
    steps = overlap.to_steps()
    assert len(steps) == 1
    assert steps.boundaries == [Range(0, 10)]
    assert steps.annotations == [set()]


def test_overlap_steps_empty_boundaries(overlap_steps_instance):
    # Create an OverlapSteps instance with no boundaries
    rng = Range(0, 10)
    boundaries = []
    annotations = []

    with pytest.raises(ValueError):
        OverlapSteps(rng, boundaries, annotations)


def test_overlap_to_steps_single_interval(overlap_instance):
    # Create an Overlap instance with a single interval
    rng = Range(0, 10)
    intervals = [Range(1, 9)]
    annotations = ['a']
    overlap = Overlap(rng, intervals, annotations)

    # Call the to_steps method
    steps = overlap.to_steps()

    # Assert that the returned OverlapSteps instance has the correct rng, boundaries, and annotations
    assert isinstance(steps, OverlapSteps)
    assert steps.rng == overlap.rng
    assert steps.boundaries == [
        Range(0, 1), Range(1, 9), Range(9, 10)
    ]
    assert steps.annotations == [set(), {'a'}, set()]



import copy
import time
from attrs import define
from attrs import field
from collections import defaultdict
from collections.abc import Iterable
from intervaltree import IntervalTree
from joblib import Parallel, delayed
from typing import TypeVar, cast

from biobit.ds.gindex import Overlap, GenomicIndex
from biobit.core import Interval, Orientation, Range
from .partition import Partition
from .reads_counter import MultiReadsCounter, CountingStats
from ..resolve import Counts, Resolution
from ..source import Source

_T = TypeVar('_T')
_K = TypeVar('_K')


def run(
        tag: _K,
        source: Source,
        partition: Partition[_T],
        resolution: Resolution[list[Overlap[_T]], Counts[_T]],
) -> tuple[_K, Counts[_T], CountingStats]:
    if len(partition) == 0:
        raise ValueError("Partition must have at least one interval")

    launched_at = time.time()

    # Build the index
    skeleton = {
        (partition.contig, Orientation.fwd): IntervalTree(),
        (partition.contig, Orientation.rev): IntervalTree(),
        (partition.contig, Orientation.dual): IntervalTree(),
    }
    assert all(i.contig == partition.contig for i in partition.intervals)

    start, end = partition.intervals[0].rng.start, partition.intervals[0].rng.end
    for interval, data in zip(partition.intervals, partition.data):
        skeleton[(partition.contig, interval.orient)].addi(
            interval.rng.start, interval.rng.end, data=data
        )

        start = min(start, interval.rng.start)
        end = max(end, interval.rng.end)

    index: GenomicIndex[_T] = GenomicIndex(skeleton)
    index_range = Range(start, end)

    # Count reads
    counts: dict[_T | None, float] = defaultdict(float)
    for blocks in source.fetch(partition.contig, index_range.start, index_range.end):
        overlaps = [index.overlap(partition.contig, blocks.orientation, rng=rng) for rng in blocks.blocks]
        for k, v in resolution(overlaps).items():
            counts[k] += v

    finished_at = time.time()

    no_overlap = counts.pop(None, 0)
    no_null_counts = cast(dict[_T, float], dict(counts))

    stats = CountingStats(
        time=finished_at - launched_at,
        partition=Interval(partition.contig, index_range),
        inside=sum(counts.values()),
        outside=no_overlap,
        extra=source.stats()
    )

    return tag, no_null_counts, stats


@define(slots=True)
class JoblibMultiReadsCounter(MultiReadsCounter[_T, _K]):
    """
    A class representing a reads counter that uses joblib to parallelize the counting process.
    """
    _sources: dict[_K, Source] = field(alias="sources")
    resolution: Resolution[list[Overlap[_T]], Counts[_T | None]]
    parallel: Parallel
    _counts: dict[_K, Counts[_T]] = field(factory=dict, init=False)
    _stats: list[CountingStats] = field(factory=list, init=False)

    def count(self, data: Iterable[_T], intervals: Iterable[Interval]):
        """
        Count the reads in the given intervals.

        :param data: The data to count.
        :param intervals: Intervals associated with the data.
        """
        _data, _intervals = list(data), list(intervals)
        if len(_data) != len(_intervals):
            raise ValueError(
                f"Data and intervals must have the same length, got {len(_data)} and {len(_intervals)} respectively."
            )

        # Group data into local partitions to support the parallel counting
        partitions = Partition.from_intervals(_data, _intervals)

        # Run the counting process
        results = self.parallel(
            delayed(run)(tag, copy.deepcopy(source), partition, self.resolution)
            for partition in partitions
            for tag, source in self._sources.items()
        )

        # Merge counts & stats
        for tag, counts, stats in results:
            _counts = self._counts.setdefault(tag, {})
            for k, v in counts.items():
                _counts[k] = _counts.get(k, 0) + v

            self._stats.append(stats)

    def counts(self) -> dict[_K, Counts[_T]]:
        return self._counts

    # TODO: I want counts as dict kind of things!!!

    def sources(self) -> dict[_K, Source]:
        return self._sources

    def stats(self) -> Iterable[CountingStats]:
        return self._stats

    def reset(self):
        self._counts.clear()
        self._stats.clear()



import sys
from pathlib import Path
from subprocess import check_call


def fetch(query: str, url: str, saveto: Path, force: bool, verbose: bool):
    if saveto.exists() and not force:
        return

    if not saveto.name.endswith(".gz"):
        raise ValueError(
            f"Biomart cache must be stored in gzip-ed TSV files: expected {saveto} to end with .gz."
        )

    saveto.unlink(missing_ok=True)
    saveto.parent.mkdir(parents=True, exist_ok=True)

    query = "".join([x.strip() for x in query.split("\n")])

    url = f"{url}/biomart/martservice?query={query}"
    ungzipped = saveto.with_suffix("")

    try:
        stdout = sys.stdout if verbose else sys.stderr
        check_call(["wget", "-O", ungzipped, url], stdout=stdout)
        check_call(["gzip", ungzipped])
    except Exception as e:
        if ungzipped.exists():
            ungzipped.unlink()
        raise e
    assert saveto.exists()



from attrs import define


@define(slots=True, frozen=True, eq=True, order=True, repr=True, hash=True)
class OverlapInfo:
    size: int
    rng: "Range"


@define(slots=True, frozen=False, eq=True, order=True, repr=True, hash=True)
class Range:
    start: int
    end: int

    def __attrs_post_init__(self):
        if self.start > self.end:
            raise ValueError(
                f"Start must be less than or equal to end. Got {self.start} > {self.end}."
            )

    def __add__(self, other):
        match other:
            case int():
                return Range(self.start + other, self.end + other)
            case _:
                raise TypeError()

    def __sub__(self, other):
        return self.__add__(-other)

    def __len__(self):
        return self.end - self.start

    def __contains__(self, item: 'Range') -> bool:
        return self.start <= item.start < item.end <= self.end

    def overlap(
            self,
            other: 'Range | None' = None,
            start: int | None = None,
            end: int | None = None,
    ) -> OverlapInfo | None:
        if other is None:
            if start is None or end is None:
                raise ValueError("Either other or start and end must be provided")
            other = Range(start, end)

        start = max(self.start, other.start)
        end = min(self.end, other.end)
        size = end - start
        if size <= 0:
            return None
        return OverlapInfo(size, Range(start, end))



from collections.abc import Iterable
from typing import TypeVar, Callable

T = TypeVar("T")


def group_within(items: list[T], *, distance: Callable[[T, T], float], maxdist: int) -> Iterable[Iterable[T]]:
    result: list[list[T]] = []
    if len(items) == 0:
        return result

    cache = [items[0]]
    for p in items[1:]:
        dist = min(distance(c, p) for c in cache)
        if dist >= maxdist:
            result.append(cache)
            cache = [p]
        else:
            cache.append(p)
    assert len(cache) >= 1
    result.append(cache)
    return result



import json
from enum import EnumMeta, Enum
from pathlib import Path
from typing import Self

from biobit.paths import CACHE
from . import gene, transcript
from .assembly import Assembly

ENSEMBL_RELEASES = {
    107: "http://jul2022.archive.ensembl.org",
    108: "http://oct2022.archive.ensembl.org",
    109: "http://feb2023.archive.ensembl.org",
    110: "http://jul2023.archive.ensembl.org",
    111: "http://jan2024.archive.ensembl.org",
}
LATEST = 111


class Loader:
    def __init__(
            self,
            name: str,
            organism: str,
            version: int,
            cache: Path | None = None,
    ):
        self._assembly = name
        self._organism = organism
        if version not in ENSEMBL_RELEASES:
            raise ValueError(
                f"Unknown Ensembl release {version} version, known releases are: {ENSEMBL_RELEASES.keys()}"
            )
        self._version = version
        self._url = ENSEMBL_RELEASES[version]
        self._cache = (
            CACHE / "ensembl" / self._organism / self._assembly / str(version)
            if cache is None
            else cache
        )
        self._cache.mkdir(parents=True, exist_ok=True)

        # Load meta information
        self._meta_path = self._cache / "meta-information.json"
        if self._meta_path.exists():
            with open(self._meta_path) as f:
                self._meta = json.load(f)
            if self._meta["schema"] != "0.1":
                raise ValueError(f"Unknown schema version {self._meta['schema']}")

            if (
                    self._meta["organism"] != organism
                    or self._meta["name"] != name
                    or self._meta["version"] != version
            ):
                raise ValueError(
                    f"Meta information mismatch: {self._meta} != {organism}, {name}, {version}. "
                    f"Please delete the cache folder ({self._cache}) and try again."
                )
        else:
            self._meta = {
                "schema": "0.1",
                "organism": organism,
                "name": name,
                "version": version,
                "cached": {
                    "transcript_attributes": [],
                    "gene_attributes": [],
                },
            }

        self.transcripts: transcript.Descriptor | None = None
        self.genes: gene.Descriptor | None = None

    def auto(self, fetch: bool = False, verbose: bool = True) -> Assembly:
        if self._version != LATEST:
            raise ValueError(
                f"Auto is supported only for the latest (version {LATEST}) GRCh38 (hsapiens) and "
                f"GRCm39 (mmusculus) Ensembl assemblies"
            )

        match (self._assembly, self._organism):
            case ("GRCh38", "hsapiens"):
                self.with_transcripts(
                    set(transcript.Attribute), fetch=fetch, verbose=verbose
                ).with_genes(set(gene.Attribute), fetch=fetch, verbose=verbose)
            case ("GRCm39", "mmusculus"):
                attributes = set(transcript.Attribute) - {
                    transcript.Attribute.MANEPlusClinical,
                    transcript.Attribute.MANESelect,
                }
                self.with_transcripts(
                    attributes, fetch=fetch, verbose=verbose
                ).with_genes(set(gene.Attribute), fetch=fetch, verbose=verbose)
            case _:
                raise ValueError(
                    f"Auto is supported only for the latest (version {LATEST}) GRCh38 (hsapiens) and "
                    f"GRCm39 (mmusculus) Ensembl assemblies"
                )
        return self.finalize()

    def with_transcripts(
            self,
            attributes: set[transcript.Attribute],
            fetch: bool = False,
            verbose: bool = True,
    ) -> Self:
        path = self._cache / "transcript-attributes.tsv.gz"
        self._load_attributes(
            path,
            "transcript_attributes",
            transcript.Attribute,
            attributes,
            fetch,
            verbose,
        )
        self.transcripts = transcript.Descriptor(attributes, path)
        return self

    def with_genes(
            self, attributes: set[gene.Attribute], fetch: bool = False, verbose: bool = True
    ) -> Self:
        path = self._cache / "gene-attributes.tsv.gz"
        self._load_attributes(
            path, "gene_attributes", gene.Attribute, attributes, fetch, verbose
        )
        self.genes = gene.Descriptor(attributes, path)
        return self

    def finalize(self) -> Assembly:
        return Assembly(
            self._assembly, self._organism, self._version, self.transcripts, self.genes
        )

    def _load_attributes(
            self,
            path: Path,
            key: str,
            enum: EnumMeta,
            attributes: set,
            fetch: bool,
            verbose: bool,
    ):
        cached: set[Enum] = {enum[x] for x in self._meta["cached"][key]}

        # Check if we need to fetch the data
        if attributes.issubset(cached):
            return

        if not fetch:
            raise ValueError(
                f"Attributes {attributes} for {key} are not available locally, please set fetch=True to download them"
            )

        # Fetch the data
        all_attributes = cached | attributes
        enum.fetch(  # type: ignore
            all_attributes,
            self._organism,
            path,
            force=True,
            url=self._url,
            verbose=verbose,
        )
        self._meta["cached"][key] = [x.name for x in all_attributes]
        self._update_meta()

    def _update_meta(self):
        with open(self._meta_path, "w") as f:
            json.dump(self._meta, f)



from . import strdeductor
from .interval import Interval
from .orientation import Orientation, OrientationLike
from .range import Range
from .strand import Strand, StrandLike, Stranded

__all__ = ["Range", "Orientation", "OrientationLike", "Strand", "StrandLike", "Stranded", "strdeductor", "Interval"]



from enum import Enum

from attrs import define, field


class Stranding(Enum):
    Unstranded = "unstranded"
    Forward = "forward"
    Reverse = "reverse"
    Unknown = "unknown"

    def __repr__(self) -> str:
        return f"Stranding({self.value})"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def normalize(cls, value: 'str | Stranding') -> 'Stranding':
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            match value.lower():
                case 'unstranded' | 'u':
                    return cls.Unstranded
                case 'forward' | 'f':
                    return cls.Forward
                case 'reverse' | 'r':
                    return cls.Reverse
                case 'unknown' | 'x':
                    return cls.Unknown
                case _:
                    raise ValueError(f"Unknown stranding: {value}")
        else:
            raise ValueError(f"Unknown stranding: {value}")


@define(slots=True, frozen=True, eq=True, order=True, hash=True, repr=True, str=True)
class Library:
    """
    A class to describe meta information about a sequencing library made from a biological sample.

    Attributes
    ----------
    source : tuple[str, ...]
        What molecules were used to generate the library?
    selection : tuple[str, ...]
        Were there any selection/enrichment steps during library generation?
    stranding : Stranding
        What is the stranding of the library?
    """
    source: tuple[str, ...] = field(converter=lambda x: tuple(x))
    selection: tuple[str, ...] = field(converter=lambda x: tuple(x))
    stranding: Stranding = field(converter=lambda x: Stranding.normalize(x))

    @source.validator
    def check_source(self, _, value):
        if not value:
            raise ValueError("Library source must be specified")

    @selection.validator
    def check_selection(self, _, value):
        if not value:
            raise ValueError("Library selection method must be specified")

    # def __repr__(self) -> str:
    #     return f"Library({self.source}, {self.selection}, {self.stranding})"
    #
    # def __str__(self) -> str:
    #     return (f"Library:\n"
    #             f"\tSource: {', '.join(self.source)}\n"
    #             f"\tSelection: {', '.join(self.selection)}\n"
    #             f"\tStranding: {self.stranding}")



from enum import Enum
from pathlib import Path
from typing import Literal

from attrs import define, field, converters


class SeqLayout(Enum):
    Paired = "paired"
    Single = "single"

    def __repr__(self) -> str:
        return f"SeqLayout({self.value})"

    def __str__(self) -> str:
        return f"{self.value}-end"

    @classmethod
    def normalize(cls, value: "SeqLayoutLike") -> 'SeqLayout':
        if isinstance(value, cls):
            return value
        elif isinstance(value, str):
            match value.lower():
                case 'paired' | 'pe':
                    return cls.Paired
                case 'single' | 'se':
                    return cls.Single
                case _:
                    raise ValueError(f"Unknown sequencing layout: {value}")
        else:
            raise ValueError(f"Unknown sequencing layout: {value}")


SeqLayoutLike = SeqLayout | Literal["paired", "single", "pe", "se"]


@define(hash=True, slots=True, frozen=True, eq=True, order=True, repr=True, str=True)
class SeqRun:
    """
    A class to represent a single run in a sequencing experiment.

    Attributes
    ----------
    ind : str
        Index of the sequencing run, should be unique within the project.
    machine : str
        Sequencing machine used for the run. E.g. 'Illumina NovaSeq 6000', 'Oxford Nanopore MinION', etc.
    layout : SeqLayout
        Layout of the sequencing run. E.g. SeqLayout.Paired, SeqLayout.Single or just 'paired', 'single'.
    files : tuple[Path, ...]
        Paths to the sequencing files. Should be one (single) or two files (paired) depending on the layout.
    reads : int
        Total number of reads in the sequencing run, if available.
    bases : int
        Total number of bases in the sequencing run, if available.
    description : str
        Description of the sequencing run, if available.
    """
    ind: str = field()
    machine: str = field()
    layout: SeqLayout = field(converter=lambda x: SeqLayout.normalize(x))
    files: tuple[Path, ...] = field(converter=lambda x: tuple(Path(f) for f in x))
    # Optional metadata
    reads: int | None = field(default=None, converter=converters.optional(lambda x: int(x)))
    bases: int | None = field(default=None, converter=converters.optional(lambda x: int(x)))
    description: str | None = field(default=None)

    @ind.validator
    def check_ind(self, _, value):
        if not value:
            raise ValueError("Sequencing run index must be specified")

    @machine.validator
    def check_machine(self, _, value):
        if not value:
            raise ValueError("Sequencing machine must be specified")

    @files.validator
    def check_files(self, _, value):
        if not value:
            raise ValueError("Sequencing files must be specified")

    @layout.validator
    def check_layout(self, _, value):
        if not value:
            raise ValueError("Sequencing layout must be specified")
        match value:
            case SeqLayout.Paired:
                if len(self.files) != 2:
                    raise ValueError("Paired-end sequencing requires two files")
            case SeqLayout.Single:
                if len(self.files) != 1:
                    raise ValueError("Single-end sequencing requires one file")
            case _:
                raise ValueError(f"Unknown sequencing layout: {value}")

    @reads.validator
    def check_reads(self, _, value):
        if value is not None and (value <= 0 or not isinstance(value, int)):
            raise ValueError("Total number of reads must be a positive integer")

    @bases.validator
    def check_bases(self, _, value):
        if value is not None and (value <= 0 or not isinstance(value, int)):
            raise ValueError("Total number of bases must be a positive integer")

    @description.validator
    def check_description(self, _, value):
        if value is not None and not value:
            raise ValueError("If specified, description must be non-empty. Use None to indicate lack of description")

    # def __repr__(self) -> str:
    #     files = ", ".join(map(str, self.files))
    #     return f"SeqRun({self.ind}, {self.machine}, {self.layout}, ({files}), {self.reads}, {self.bases}, {self.description})"
    #
    # def __str__(self) -> str:
    #     fields = [
    #         f"\tMachine: {self.machine}",
    #         f"\tLayout: {self.layout}",
    #         f"\tFiles: {', '.join(map(str, self.files))}",
    #         f"\tReads: {self.reads if self.reads else '.'}",
    #         f"\tBases: {self.bases if self.bases else '.'}",
    #         f"\tDescription: {self.description if self.description else '.'}"
    #     ]
    #     body = "\n".join(fields)
    #     return f"SeqRun({self.ind}):\n{body}"



from typing import Iterator, Protocol

from attrs import define

from biobit.core import Range, Orientation


@define(slots=True, frozen=True)
class AlignedBlocks:
    blocks: tuple[Range, ...]
    orientation: Orientation


class Source(Protocol):
    """
    Abstraction that represents a source of reads. The source can be directed to fetch reads from a specific region
    and then iterated over to yield the reads in that region. Each read comes as a tuple of strand and a list of
    blocks (ranges) that denote the read's alignment to the reference genome.

    Sources are though to be cheap to clone, but not to be shared between threads.
    """

    def fetch(self, contig: str, start: int, end: int) -> Iterator[AlignedBlocks]:
        """
        Fetch reads from a specific region of the reference genome.
        :param contig: The contig to fetch reads from.
        :param start: The start position of the region.
        :param end: The end position of the region.
        :return: Iterator of read blocks in the region.
        """
        pass

    def stats(self) -> dict:
        """
        Return statistics about processed reads.
        :return: A dictionary containing statistics about the source. The content is implementation-specific.
        """
        pass



from collections import defaultdict
from pathlib import Path
from typing import Iterable, Self, Iterator

from attrs import define, field
from pysam import AlignedSegment, AlignmentFile


@define(slots=True)
class ConsumedReads:
    paired: int = 0
    proper_pair: int = 0
    unmapped: int = 0
    mate_unmapped: int = 0
    reverse_strand: int = 0
    mate_reverse_strand: int = 0
    first_in_pair: int = 0
    second_in_pair: int = 0
    not_primary_alignment: int = 0
    fails_quality_checks: int = 0
    duplicate: int = 0
    supplementary: int = 0
    total: int = 0

    def count(self, read: AlignedSegment):
        if read.is_paired:
            self.paired += 1
        if read.is_proper_pair:
            self.proper_pair += 1
        if read.is_unmapped:
            self.unmapped += 1
        if read.mate_is_unmapped:
            self.mate_unmapped += 1
        if read.is_reverse:
            self.reverse_strand += 1
        if read.mate_is_reverse:
            self.mate_reverse_strand += 1
        if read.is_read1:
            self.first_in_pair += 1
        if read.is_read2:
            self.second_in_pair += 1
        if read.is_secondary:
            self.not_primary_alignment += 1
        if read.is_qcfail:
            self.fails_quality_checks += 1
        if read.is_duplicate:
            self.duplicate += 1
        if read.is_supplementary:
            self.supplementary += 1
        self.total += 1

    def summarize(self, prefix: str = "") -> str:
        return (
            f"{prefix}Paired: {self.paired}({self.paired / self.total:.2%})\n"
            f"{prefix}Proper pair: {self.proper_pair}({self.proper_pair / self.total:.2%})\n"
            f"{prefix}Unmapped: {self.unmapped}({self.unmapped / self.total:.2%})\n"
            f"{prefix}Mate unmapped: {self.mate_unmapped}({self.mate_unmapped / self.total:.2%})\n"
            f"{prefix}Reverse strand: {self.reverse_strand}({self.reverse_strand / self.total:.2%})\n"
            f"{prefix}Mate reverse strand: {self.mate_reverse_strand}({self.mate_reverse_strand / self.total:.2%})\n"
            f"{prefix}First in pair: {self.first_in_pair}({self.first_in_pair / self.total:.2%})\n"
            f"{prefix}Second in pair: {self.second_in_pair}({self.second_in_pair / self.total:.2%})\n"
            f"{prefix}Not primary alignment: {self.not_primary_alignment}({self.not_primary_alignment / self.total:.2%})\n"
            f"{prefix}Fails quality checks: {self.fails_quality_checks}({self.fails_quality_checks / self.total:.2%})\n"
            f"{prefix}Duplicate: {self.duplicate}({self.duplicate / self.total:.2%})\n"
            f"{prefix}Supplementary: {self.supplementary}({self.supplementary / self.total:.2%})\n"
            f"{prefix}Total: {self.total}\n"
        )

    def __str__(self):
        return self.summarize()


class Reader:
    def __init__(self, filename: Path, inflags: int, exflags: int, minmapq: int, *, statistics: bool = False):
        self.filename: Path = filename
        self.sf: AlignmentFile | None = None
        self.iterator: Iterable[AlignedSegment] | None = None
        self.inflags: int = inflags
        self.exflags: int = exflags
        self.minmapq: int = minmapq

        self.statistics: bool = statistics
        self.consumed: ConsumedReads | None = None if not statistics else ConsumedReads()
        self.discarded: ConsumedReads | None = None if not statistics else ConsumedReads()

    def fetch(self, contig: str, start: int | None = None, end: int | None = None) -> Self:
        if self.sf is None:
            self.sf = AlignmentFile(self.filename.as_posix(), "rb")
        self.iterator = self.sf.fetch(contig, start, end)
        return self

    def _is_read_ok(self, segment: AlignedSegment) -> bool:
        return (
                segment.flag & self.inflags == self.inflags
                and segment.flag & self.exflags == 0
                and segment.mapping_quality >= self.minmapq
        )

    def _iter_with_statistics(self) -> Iterator[AlignedSegment]:
        assert self.consumed is not None and self.discarded is not None
        assert self.iterator is not None

        for segment in self.iterator:
            if self._is_read_ok(segment):
                self.consumed.count(segment)
                yield segment
            else:
                self.discarded.count(segment)

        self.iterator = None

    def _iter_without_statistics(self) -> Iterator[AlignedSegment]:
        assert self.iterator is not None
        for segment in self.iterator:
            if self._is_read_ok(segment):
                yield segment

        self.iterator = None

    def __iter__(self) -> Iterator[AlignedSegment]:
        if self.sf is None:
            self.sf = AlignmentFile(self.filename.as_posix(), "rb")
            self.iterator = self.sf

        if self.iterator is None:
            self.iterator = self.sf

        if self.statistics:
            return self._iter_with_statistics()
        else:
            return self._iter_without_statistics()

    def __deepcopy__(self, _):
        return Reader(self.filename, self.inflags, self.exflags, self.minmapq, statistics=self.statistics)

    def __getstate__(self):
        return self.filename, self.inflags, self.exflags, self.minmapq, self.statistics

    def __setstate__(self, state):
        self.filename, self.inflags, self.exflags, self.minmapq, self.statistics = state
        self.sf = None
        self.consumed = None if not self.statistics else ConsumedReads()
        self.discarded = None if not self.statistics else ConsumedReads()


@define(slots=True)
class _PEReadsCache:
    lmates: list[AlignedSegment] = field(factory=list)
    rmates: list[AlignedSegment] = field(factory=list)

    def bundle(self) -> Iterator[tuple[AlignedSegment, AlignedSegment]]:
        missed_lmates = []
        for lmate in self.lmates:
            mate_found = False
            for ind, rmate in enumerate(self.rmates):
                if (
                        lmate.next_reference_id == rmate.reference_id
                        and lmate.next_reference_start == rmate.reference_start
                        and lmate.mate_is_reverse == rmate.is_reverse
                        and lmate.mate_is_unmapped == rmate.is_unmapped
                        and rmate.next_reference_id == lmate.reference_id
                        and rmate.next_reference_start == lmate.reference_start
                        and rmate.mate_is_reverse == lmate.is_reverse
                        and rmate.mate_is_unmapped == lmate.is_unmapped
                ):
                    mate_found = True
                    break

            if not mate_found:
                missed_lmates.append(lmate)
            else:
                rmate = self.rmates.pop(ind)
                yield lmate, rmate

        self.lmates = missed_lmates

    def is_empty(self) -> bool:
        return len(self.lmates) == 0 and len(self.rmates) == 0


class PEReadsBundler:
    def __init__(self, reader: Reader, step: int = 100_000):
        self.reader: Reader = reader
        self.cache: dict[str, _PEReadsCache] = defaultdict(_PEReadsCache)

        self.step: int = step
        self.unpaired = 0

    def fetch(self, contig: str, start: int | None = None, end: int | None = None) -> Self:
        self.reader.fetch(contig, start, end)
        return self

    def __iter__(self) -> Iterator[tuple[AlignedSegment, AlignedSegment]]:
        for ind, segment in enumerate(self.reader):
            qname = segment.query_name
            assert qname is not None, "Query name must be present in the BAM file"

            cached = self.cache[qname]
            if segment.is_read1:
                cached.lmates.append(segment)
            else:
                cached.rmates.append(segment)

            if ind % self.step == 0 and ind > 0:
                for k, v in list(self.cache.items()):
                    yield from v.bundle()
                    if v.is_empty():
                        self.cache.pop(k)

        # Clean up the cache
        for k, v in list(self.cache.items()):
            yield from v.bundle()
            self.unpaired += len(v.lmates) + len(v.rmates)
        self.cache.clear()

    def summarize(self, prefix: str = "") -> str:
        return f"{prefix}Unpaired reads left: {self.unpaired}\n"



from collections.abc import Callable

from biobit.analysis import seqproj

__all__ = ["from_bioexp", "to_expind"]


def from_bioexp(
        experiment: seqproj.Experiment, *,
        title_builder: Callable[[seqproj.Experiment], str] | str | None = None,
        separator: str = "+"
) -> str:
    """
    Converts a bioproj experiment into a human-readable descriptor that can be used for
    nf-core/rnaseq pipeline design files.

    :param experiment: The bioproj.Experiment object to be converted.
    :param title_builder: A function that converts a bioproj.Experiment object into a human-readable title. If string,
        the title is set to the attribute with the given name. If None, the title is built by concatenating all
        experiment sample attributes. If the sample has no attributes, the sample description is used as the title.
    :param separator: The string used to separate the experiment index and its title in the resulting descriptor.
    :return: A string descriptor, composed of the experiment index and its title, separated by the specified separator.
    """
    if callable(title_builder):
        title = title_builder(experiment)
    elif isinstance(title_builder, str):
        if title_builder not in experiment.sample.attributes:
            raise ValueError(f"Attribute '{title_builder}' not found in the sample attributes: {experiment.sample}")
        title = experiment.sample.attributes[title_builder]
    else:
        attributes: list[str] = []
        for atrribute, value in sorted(experiment.sample.attributes.items()):
            attributes.append(f"{atrribute}={value}")

        # If no tags are present, use the sample description as a descriptor
        if len(attributes) == 0:
            if experiment.sample.description is None:
                raise ValueError("No tags or description available for the sample.")
            attributes.append(experiment.sample.description)

        title = "_".join(attributes)

    descriptor = f"{experiment.ind}{separator}{title}"
    for char in [",", "|", "/", "[", "]"]:
        if char in descriptor:
            raise ValueError(f"Descriptor '{descriptor}' contains prohibited character '{char}'")

    return descriptor


def to_expind(descriptor: str, separator: str = "+") -> str:
    """
    Extracts the experiment index from a descriptor string.

    :param descriptor: The descriptor string from which to extract the experiment index.
    :param separator: The string used to separate the experiment index and the title in the descriptor.
    :return: A string representing the experiment index.
    """
    return descriptor.split(separator)[0]



from attrs import define, field


@define(hash=True, slots=True, frozen=True, eq=True, order=True, repr=True, str=True)
class Sample:
    """
    A class to represent a sample in a biological experiment.

    Attributes
    ----------
    ind : str
        Sample ID, must be unique within the project.
    organism : tuple[str, ...]
        Organism(s) from which the sample was derived. E.g. ('Homo sapiens',), ('Mus musculus', 'HSV-1'), etc.
    attributes: dict[str, str]
        Additional descriptive attributes for the sample, optional. E.g. {'Confluence': '75%', 'Source': 'HeLa'}, etc.
    description: str
        A description of the sample, if available.
    """
    ind: str = field()
    organism: tuple[str, ...] = field(converter=lambda x: tuple(x))
    # Optional meta information
    attributes: dict[str, str] = field(factory=dict)
    description: str | None = field(default=None)

    @ind.validator
    def check_ind(self, _, value):
        if not value:
            raise ValueError("Sample ID must be specified")

    @organism.validator
    def check_organism(self, _, value):
        if not value:
            raise ValueError("Organism must be specified")

    @attributes.validator
    def check_tags(self, _, value):
        for k, v in value.items():
            if not v:
                raise ValueError(f"Empty attributes are not allowed: {k} -> {v}")

    # def __repr__(self) -> str:
    #     return f"Sample({self.ind}, {self.organism}, {self.attributes}, {self.description})"
    #
    # def __str__(self) -> str:
    #     return (f"Sample({self.ind}):\n"
    #             f"\tOrganism: {', '.join(self.organism)}\n"
    #             f"\tAttributes: {', '.join(f'{k}={v}' for k, v in self.attributes.items())}\n"
    #             f"\tDescription: {self.description or '.'}")



from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable
from typing import TypeVar, Generic, cast

from biobit.ds.gindex import Overlap

__all__ = ["Resolution", "Proportional", "Binary", "Priority", "Normalize", "Counts"]

_T = TypeVar("_T")
_I = TypeVar("_I")
_O = TypeVar("_O")
_D = TypeVar("_D")

Counts = dict[_T, float]


class Resolution(Generic[_I, _O], ABC):
    # Resolution is required to turn a list of overlapping annotations into weights for one or more annotation categories.
    # This is necessary when a sequenced read/read pair isn't entirely within a known annotation 'type'.
    # All resolution strategies are chainable, so that multiple strategies can be applied in sequence.
    @abstractmethod
    def __call__(self, data: _I) -> _O:
        pass

    def apply(self, nxt: 'Callable[[_O], _T] | Resolution[_O, _T]') -> 'Resolution[_I, _T]':
        if isinstance(nxt, Resolution):
            return _Chain(self, nxt)
        elif isinstance(nxt, Callable):  # type: ignore # https://github.com/python/mypy/issues/6680
            return _Chain(self, _Lambda(nxt))
        else:
            raise TypeError(f"Cannot apply {nxt} to {self}")

    def __or__(self, nxt: 'Callable[[_O], _T] | Resolution[_O, _T]') -> 'Resolution[_I, _T]':
        return self.apply(nxt)


class _Chain(Generic[_I, _O, _T], Resolution[_I, _T]):
    def __init__(self, first: Resolution[_I, _O], second: Resolution[_O, _T]):
        self.first = first
        self.second = second

    def __call__(self, data: _I) -> _T:
        return self.second(self.first(data))


class _Lambda(Generic[_I, _O, _T], Resolution[_I, _O]):
    def __init__(self, fn: Callable[[_I], _O]):
        self.fn = fn

    def __call__(self, data: _I) -> _O:
        return self.fn(data)


class Proportional(Resolution[list[Overlap[_T]], Counts[_T | None]]):
    def __init__(self, empty: _T | None = None):
        self.empty = empty

    def __call__(self, overlaps: list[Overlap[_T]]) -> Counts[_T | None]:
        """
        The proportional strategy assigns weights to annotation categories based on the overlap of the target ROI
        with each category
        """
        if len(overlaps) == 0:
            return {self.empty: 1}

        weights: Counts[_T | None] = {}
        covered, total = 0, 0
        for roiov in overlaps:
            total += len(roiov.rng)  # original ROI length
            # step-wise ROI annotation
            for rng, annotations in roiov.to_steps():  # type: ignore # https://github.com/python/mypy/issues/16733
                length = len(rng)
                if not annotations:
                    covered += length
                    weights[self.empty] = weights.get(self.empty, 0) + length
                else:
                    covered += length
                    for anno in annotations:
                        weights[anno] = weights.get(anno, 0) + length

        assert covered == total, f"Covered length {covered} does not match total length {total}"
        return weights


class Binary(Resolution[list[Overlap[_T]], Counts[_T | None]]):
    def __init__(self, empty: _T | None = None):
        self.empty = empty

    def __call__(self, overlaps: list[Overlap[_T]]) -> Counts[_T | None]:
        """
        The binary strategy assigns weights to annotation categories based on the presence of at least a single base pair
        overlap between the read block and each category
        """
        categories: set[_T | None] = set()
        length, covered = 0, 0
        for blockov in overlaps:
            length += len(blockov.rng)
            for rng, step in blockov.to_steps():  # type: ignore # https://github.com/python/mypy/issues/16733
                if step:
                    covered += len(rng)
                    categories |= step

        if not categories:
            return {self.empty: 1}

        if covered != length:
            categories.add(self.empty)

        weights = {cat: 1.0 for cat in categories}
        return weights


class Priority(Generic[_T, _D], Resolution[Counts[_T], Counts[_T]]):
    def __init__(self, priority: list[_D], key: Callable[[_T], _D] | None = None):
        self.priority = priority
        self.key = key or (lambda x: cast(_D, x))

    def __call__(self, weights: Counts[_T]) -> Counts[_T]:
        """
        The priority strategy picks the first category in the priority list that is also present in the weights.
        If more than one category is mapped to the same priority, all such categories are returned.
        """
        keys: defaultdict[_D, list[_T]] = defaultdict(list)
        for cat in weights:
            keys[self.key(cat)].append(cat)

        categories: list[_T] | None = None
        for p in self.priority:
            if p in keys:
                categories = keys[p]
                break

        if categories is None:
            raise ValueError(f"None of the priority categories {self.priority} are present in the weights {weights}")

        return {cat: weights[cat] for cat in categories}


class Normalize(Resolution[Counts[_T], Counts[_T]]):
    def __init__(self, total: float = 1.0):
        self.total = total

    def __call__(self, weights: Counts[_T]) -> Counts[_T]:
        """
        Normalize annotation weights to sum to a given total
        """
        if len(weights) == 0:
            return weights

        total = sum(weights.values())
        if total == 0:
            raise ValueError(f"Total weight is 0, cannot normalize {weights} to {self.total}")

        factor = self.total / total
        return {k: v * factor for k, v in weights.items()}



import os
from itertools import chain
from pathlib import Path
from typing import Iterable

from attrs import define, field

from biobit.analysis import seqproj
from biobit.core import Stranded


@define(slots=True, frozen=True, eq=True, order=True, repr=True, hash=True, init=False)
class Experiment:
    descriptor: str = field()
    source: seqproj.Experiment = field()
    # Files generated by the pipeline
    bam: Path = field(converter=lambda x: Path(x))
    bigwig: Stranded[Path] = field(converter=lambda x: Stranded(*[Path(p) for p in x]))
    salmon: Path = field(converter=lambda x: Path(x))

    def __init__(self, descriptor: str, source: seqproj.Experiment, resfolder: os.PathLike[str]):
        # Folder with the results of the pipeline
        resfolder = Path(resfolder)
        # if not resfolder.is_dir():
        #     raise ValueError(f"Results folder {resfolder} does not exist")

        # BAM file
        bam = resfolder / "star_salmon" / f"{descriptor}.markdup.sorted.bam"

        # Bigwig
        if source.library.stranding in {seqproj.Stranding.Unstranded, seqproj.Stranding.Unknown}:
            raise ValueError(
                f"Unstranded experiments or experiments with unknown strandedness are not supported: {descriptor}"
            )

        fwd = resfolder / "star_salmon" / "bigwig" / f"{descriptor}.forward.bigWig"
        rev = resfolder / "star_salmon" / "bigwig" / f"{descriptor}.reverse.bigWig"

        bigwig = Stranded(fwd, rev)

        # Salmon file
        salmon = resfolder / "star_salmon" / descriptor / "quant.sf"

        self.__attrs_init__(descriptor, source, bam, bigwig, salmon)

    @property
    def ind(self) -> str:
        return self.source.ind

    @property
    def library(self) -> seqproj.Library:
        return self.source.library

    @property
    def sample(self) -> seqproj.Sample:
        return self.source.sample

    @property
    def runs(self) -> Iterable[seqproj.SeqRun]:
        return self.source.runs

    @property
    def ensure_exists(self):
        allfiles = chain(
            [self.bam, self.bam.with_suffix(".bam.bai"), self.bigwig.fwd, self.bigwig.rev, self.salmon],
            *[r.files for r in self.runs]
        )
        missing = [f for f in allfiles if not f.exists()]
        if missing:
            report = "\n".join(f"\t{f}" for f in missing)
            raise FileNotFoundError(
                f"nf-core/rnaseq project ({self.source.ind}) is not complete, missing files: {report}"
            )



from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path
from typing import Callable, TypeVar, Generic

from intervaltree import IntervalTree
from pybedtools import BedTool, Interval as BedInterval

from biobit.core import Range, Orientation, OrientationLike, Interval
from .overlap import Overlap

_T = TypeVar("_T")


def bedname(it: BedInterval) -> str:
    return it.name


def merge(*indices: "GenomicIndex[_T]") -> "GenomicIndex[_T]":
    if not indices:
        raise ValueError("Can't merge empty list of genomic indexes")
    elif len(indices) == 1:
        return indices[0]

    keys: set[tuple[str, Orientation]] = set()
    for ind in indices:
        for key in ind.itrees.keys():
            keys.add(key)

    itrees = {}
    for key in keys:
        merged = set()
        for ind in indices:
            tree = ind.itrees.get(key, None)
            if tree is not None:
                merged |= set(tree)

        itrees[key] = IntervalTree(merged)
    return GenomicIndex(itrees)


class GenomicIndex(Generic[_T]):
    itrees: dict[tuple[str, Orientation], IntervalTree]

    def __init__(self, itrees: dict[tuple[str, Orientation], IntervalTree] | None = None):
        self.itrees = itrees if itrees else {}

    def set(self, contig: str, orientation: OrientationLike, index: IntervalTree):
        self.itrees[(contig, Orientation.normalize(orientation))] = index

    def overlap(
            self,
            contig: str,
            orientation: OrientationLike,
            start: int | None = None,
            end: int | None = None,
            rng: Range | None = None,
    ) -> Overlap[_T]:
        orient = Orientation.normalize(orientation)

        if (start is None or end is None) and rng is None:
            raise ValueError("Either start and end or range must be provided")
        elif (start is not None or end is not None) and rng is not None:
            raise ValueError("Either start and end or range must be provided, not both")

        if rng is None:
            if start is None or end is None:
                raise ValueError("Both start and end must be provided")
            rng = Range(start, end)

        index = self.itrees.get((contig, orient), None)
        if index is None:
            return Overlap(rng, [], [])

        hits, annotation = [], []
        for it in sorted(index.overlap(rng.start, rng.end), key=lambda x: x.begin):
            ov = rng.overlap(start=it.begin, end=it.end)
            assert ov is not None
            hits.append(ov.rng)
            annotation.append(it.data)
        return Overlap(rng, hits, annotation)

    def merge(self, *indices: 'GenomicIndex[_T]', inplace: bool = False) -> 'GenomicIndex[_T]':
        merged = merge(self, *indices)
        if inplace:
            self.itrees = merged.itrees
            return self
        else:
            return merged

    def subset(
            self, contig: str, start: int, end: int, orientation: OrientationLike | None = None
    ) -> "GenomicIndex[_T]":
        if orientation is None:
            orients = [Orientation.fwd, Orientation.rev, Orientation.dual]
        else:
            orients = [Orientation.normalize(orientation)]

        trees = {}
        for orient in orients:
            index = self.itrees.get((contig, orient), None)
            if index is not None:
                tree = IntervalTree()
                for it in index.overlap(start, end):
                    tree.addi(it.begin, it.end, data=it.data)
                trees[(contig, orient)] = tree

        return GenomicIndex(trees)

    def covered(self) -> Iterable[Interval]:
        for (contig, orientation), tree in self.itrees.items():
            for it in tree.all_intervals:
                yield Interval(contig, Range(it.begin, it.end), orientation)

    @staticmethod
    def from_bed(
            bed: Path,
            datafn: Callable[[BedInterval], _T] = bedname,  # type: ignore
            filterfn: Callable[[BedInterval], bool] | None = None,
    ) -> "GenomicIndex[_T]":
        filterfn = filterfn if filterfn is not None else (lambda _: True)

        itrees: defaultdict[tuple[str, Orientation], IntervalTree] = defaultdict(IntervalTree)
        for it in BedTool(bed):
            if not filterfn(it):
                continue
            key = (it.chrom, Orientation.normalize(it.strand))
            itrees[key].addi(it.start, it.end, data=datafn(it))
        result: dict[tuple[str, Orientation], IntervalTree] = dict(itrees)
        return GenomicIndex(result)



from collections.abc import Iterable
from dataclasses import dataclass
from typing import Generic, TypeVar

from sortedcontainers import SortedList

from biobit.core import Range

_T = TypeVar("_T")


@dataclass(frozen=True, slots=True)
class OverlapSteps(Generic[_T]):
    rng: Range
    boundaries: list[Range]
    annotations: list[set[_T]]

    def __post_init__(self):
        if len(self.boundaries) != len(self.annotations):
            raise ValueError("Boundaries and annotations must have the same length")
        if len(self.boundaries) == 0:
            raise ValueError("Boundaries and annotations must have at least one element")
        if self.boundaries[0].start != self.rng.start or self.boundaries[-1].end != self.rng.end:
            raise ValueError("Boundaries must start at the beginning of the range and end at the end of the range")

    def __len__(self) -> int:
        return len(self.annotations)

    def __iter__(self) -> Iterable[tuple[Range, set[_T]]]:
        return zip(self.boundaries, self.annotations)


@dataclass(frozen=True, slots=True)
class Overlap(Generic[_T]):
    rng: Range
    intervals: list[Range]
    annotations: list[_T]

    def to_steps(self) -> OverlapSteps[_T]:
        template = {self.rng.start, self.rng.end}
        for it in self.intervals:
            template.add(it.start)
            template.add(it.end)

        boundaries = SortedList(template)

        annotation: list[set[_T]] = [set() for _ in range(len(boundaries) - 1)]

        for it, anno in zip(self.intervals, self.annotations):
            st, en = boundaries.bisect_left(it.start), boundaries.bisect_left(it.end)
            for stanno in annotation[st:en]:
                stanno.add(anno)
        steps = [Range(boundaries[i], boundaries[i + 1]) for i in range(len(boundaries) - 1)]

        return OverlapSteps(self.rng, steps, annotation)

    def __iter__(self) -> Iterable[tuple[Range, _T]]:
        return zip(self.intervals, self.annotations)

    def __len__(self) -> int:
        return len(self.intervals)



from typing import Protocol, TypeVar, Iterable

from attrs import define, field

from biobit.core import Interval
from ..resolve import Counts
from ..source import Source

_T = TypeVar('_T')
_K = TypeVar('_K')


@define(slots=True, frozen=True)
class CountingStats:
    time: float
    partition: Interval
    inside: float
    outside: float

    extra: dict[str, float] = field(factory=dict)

    def to_dict(self) -> dict:
        return {
            "Time(s)": self.time,
            "Partition": f"{self.partition.contig}:{self.partition.rng.start}-{self.partition.rng.end}",
            "Inside ROIs": self.inside,
            "Outside ROIs": self.outside,
            **self.extra,
        }


class MultiReadsCounter(Protocol[_T, _K]):
    def count(self, data: Iterable[_T], intervals: Iterable[Interval]):
        ...

    def counts(self) -> dict[_K, Counts[_T]]:
        ...

    def sources(self) -> dict[_K, Source]:
        ...

    def stats(self) -> Iterable[CountingStats]:
        ...

    def reset(self):
        ...



import pytest

from biobit.ds.gindex import Overlap
from biobit.core import Range
from .resolve import _Chain, _Lambda, Proportional, Binary, Priority, Normalize


# Test ResolutionChain
def test_resolution_chain():
    # Define two simple resolutions
    res1 = _Lambda(lambda x: x + 1)
    res2 = _Lambda(lambda x: x * 2)

    # Chain them
    chain = _Chain(res1, res2)

    # Test
    assert chain(2) == 6  # (2 + 1) * 2
    assert res1.apply(res2)(2) == 6
    assert (res1 | res2)(2) == 6


# Test ResolutionLambda
def test_resolution_lambda():
    # Define a simple resolution
    res = _Lambda(lambda x: x + 1)

    def fn(x):
        return x * 2

    # Test
    assert res(2) == 3
    assert res.apply(fn)(2) == 6
    assert (res | fn)(2) == 6


# Test Proportional
def test_proportional():
    # Define a simple proportional resolution
    prop = Proportional(empty="None")

    # Test with no overlaps
    assert prop([]) == {"None": 1}
    assert prop([Overlap(Range(0, 10), [], [])]) == {"None": 10}

    # Test with a single overlap
    overlaps = [Overlap(Range(0, 10), [Range(0, 7)], ['a'])]
    assert prop(overlaps) == {"a": 7, "None": 3}

    overlaps = [Overlap(Range(0, 7), [Range(0, 7)], ['a'])]
    assert prop(overlaps) == {"a": 7}

    # Test with several overlaps
    overlaps = [
        Overlap(Range(0, 10), [Range(0, 7), Range(3, 9)], ['a', 'b']),
        Overlap(Range(11, 15), [Range(14, 15)], ['a']),
    ]
    weights = {"a": 8, "b": 6, "None": 4}
    assert prop(overlaps) == weights


def test_binary():
    binary = Binary(empty=None)

    # Test with no overlaps
    assert binary([]) == {None: 1}
    assert binary([Overlap(Range(0, 10), [], [])]) == {None: 1}

    # Test with a single overlap
    overlaps = [Overlap(Range(0, 10), [Range(0, 7)], ['a'])]
    assert binary(overlaps) == {"a": 1, None: 1}

    overlaps = [Overlap(Range(0, 7), [Range(0, 7)], ['a'])]
    assert binary(overlaps) == {"a": 1}

    # Test with several overlaps
    overlaps = [
        Overlap(Range(0, 10), [Range(0, 7), Range(3, 9)], ['a', 'b']),
        Overlap(Range(11, 15), [Range(14, 15)], ['a']),
    ]
    assert binary(overlaps) == {"a": 1, "b": 1, None: 1}


def test_priority():
    priority = Priority(['a', 'b'])

    # Test unknown category
    for weights in [{}, {'c': 1}]:
        with pytest.raises(ValueError):
            priority(weights)

    # Test with a single category
    assert priority({'a': 1}) == {'a': 1}

    # Test with several categories
    weights = {'a': 0.5, 'b': 0.5}
    assert priority(weights) == {'a': 0.5}


def test_normalize():
    normalize = Normalize(total=10)

    # Test with no weights
    assert normalize({}) == {}

    # Test with a single weight
    assert normalize({'a': 12}) == {'a': 10}
    assert normalize({'a': 10}) == {'a': 10}

    # Special case
    with pytest.raises(ValueError):
        normalize({'a': 0})

    # Test with several weights
    weights = {'a': 50, 'b': 25, 'c': 25}
    assert normalize(weights) == {'a': 5, 'b': 2.5, 'c': 2.5}

    normalize.total = 1
    weights = {'a': 1, 'b': 2, 'c': 3}
    assert normalize(weights) == {'a': 1 / 6, 'b': 2 / 6, 'c': 3 / 6}


def test_priority_with_key():
    priority = Priority(['a', 'b', 'c'], key=lambda x: x[0])

    # Test unknown category
    for weights in [{}, {'da': 1}]:
        with pytest.raises(ValueError):
            priority(weights)

    # Test with a single category
    assert priority({'a': 1}) == {'a': 1}
    assert priority({('a', 'b'): 1}) == {('a', 'b'): 1}
    assert priority({'ba': 1, 'bc': 2}) == {'ba': 1, 'bc': 2}

    # Test with several categories
    assert priority({('a', 1): 0.5, ('b', 2): 1}) == {('a', 1): 0.5}
    assert priority({'bac': 1, 'ca': 2}) == {'bac': 1}
    assert priority({'bac': 1, 'ca': 2, 'b': 3}) == {'bac': 1, 'b': 3}



from abc import ABC
from enum import Enum
from pathlib import Path
from typing import Self


class EnsemblAttribute(Enum):
    @classmethod
    def fetch(
            cls,
            attributes: set[Self],
            organism: str,
            saveto: Path,
            force: bool,
            url: str,
            verbose: bool,
    ):
        raise ValueError("Not implemented")

    @property
    def colname(self) -> str:
        raise ValueError("Not implemented")

    @property
    def dtype(self):
        raise ValueError("Not implemented")


class EnsemblDescriptor(ABC):
    def _getter(self, mapping, id):
        if id not in mapping:
            id = id.split(".")[0]
        return mapping.get(id, None)



import pytest

from biobit.core import strdeductor, Strand


def test_get():
    assert strdeductor.get("f/s") == strdeductor._fs
    assert strdeductor.get("s/f") == strdeductor._sf
    assert strdeductor.get("f") == strdeductor._f
    assert strdeductor.get("s") == strdeductor._s
    with pytest.raises(ValueError):
        strdeductor.get("invalid")  # type: ignore


def test_f():
    assert strdeductor._f(_AlignedSegmentStub(True)) == Strand.fwd
    assert strdeductor._f(_AlignedSegmentStub(False)) == Strand.rev


def test_s():
    assert strdeductor._s(_AlignedSegmentStub(True)) == Strand.rev
    assert strdeductor._s(_AlignedSegmentStub(False)) == Strand.fwd


def test_fs():
    assert strdeductor._fs(_AlignedSegmentStub(True, True)) == Strand.fwd
    assert strdeductor._fs(_AlignedSegmentStub(True, False)) == Strand.rev
    assert strdeductor._fs(_AlignedSegmentStub(False, True)) == Strand.rev
    assert strdeductor._fs(_AlignedSegmentStub(False, False)) == Strand.fwd


def test_sf():
    assert strdeductor._sf(_AlignedSegmentStub(True, True)) == Strand.rev
    assert strdeductor._sf(_AlignedSegmentStub(True, False)) == Strand.fwd
    assert strdeductor._sf(_AlignedSegmentStub(False, True)) == Strand.fwd
    assert strdeductor._sf(_AlignedSegmentStub(False, False)) == Strand.rev


class _AlignedSegmentStub:
    def __init__(self, is_reverse, is_read1=None):
        self.is_reverse = is_reverse
        self.is_read1 = is_read1



from . import adapter
from .experiment import Experiment
from .library import Library, Stranding
from .project import Project
from .sample import Sample
from .seqrun import SeqLayout, SeqRun

__all__ = ["adapter", "Experiment", "Library", "Stranding", "Project", "Sample", "SeqLayout", "SeqRun"]



import intervaltree as it
import pytest

from biobit.ds.gindex import GenomicIndex
from biobit.core import Orientation, Range


@pytest.fixture
def empty_index():
    return GenomicIndex()


@pytest.fixture
def complex_index():
    itrees = {}

    cache = it.IntervalTree()
    cache.addi(1, 10, data='1:10')
    cache.addi(20, 30, data='20:30')
    itrees[('1', Orientation.fwd)] = cache

    cache = it.IntervalTree()
    cache.addi(-5, 2, data='-5:2')
    cache.addi(0, 150, data='0:150')
    cache.addi(200, 300, data='200:300')
    itrees[('2', Orientation.fwd)] = cache

    cache = it.IntervalTree()
    cache.addi(0, 5, data='0:5')
    cache.addi(3, 10, data='3:10')
    itrees[('2', Orientation.rev)] = cache

    return GenomicIndex(itrees)


@pytest.fixture
def simple_index():
    cache = it.IntervalTree()
    cache.addi(-100, 100, data='-100:100')
    itrees = {('1', Orientation.fwd): cache}
    return GenomicIndex(itrees)


def test_merge(empty_index, simple_index, complex_index):
    merged = GenomicIndex.merge(empty_index, simple_index, complex_index)
    assert len(merged.itrees) == 3
    assert len(merged.itrees[('1', Orientation.fwd)]) == 3
    assert len(merged.itrees[('2', Orientation.fwd)]) == 3
    assert len(merged.itrees[('2', Orientation.rev)]) == 2

    overlap = merged.overlap('1', Orientation.fwd, 0, 10)
    assert overlap.intervals == [Range(0, 10), Range(1, 10)]
    assert overlap.annotations == ['-100:100', '1:10']

    overlap = merged.overlap('2', Orientation.fwd, -3, 10)
    assert overlap.intervals == [Range(-3, 2), Range(0, 10)]
    assert overlap.annotations == ['-5:2', '0:150']

    overlap = merged.overlap('2', Orientation.rev, -1, 8)
    assert overlap.intervals == [Range(0, 5), Range(3, 8)]
    assert overlap.annotations == ['0:5', '3:10']


def test_overlap_both_range_and_start_end_provided(empty_index):
    with pytest.raises(ValueError):
        empty_index.overlap('1', Orientation.fwd, 0, 10, Range(0, 10))


def test_overlap_no_range_provided(empty_index):
    with pytest.raises(ValueError):
        empty_index.overlap('1', Orientation.fwd)


def test_overlap_no_hits(complex_index):
    overlap = complex_index.overlap('1', Orientation.fwd, 100, 200)
    assert overlap.intervals == []
    assert overlap.annotations == []


@pytest.fixture
def bed_file(tmp_path):
    data = """\
chr1	0	1000	gene1	0	+
chr1	1000	2000	gene2	0	-
chr2	0	500	gene3	0	+
chr2	500	1000	gene4	0	-
"""
    bed_file_path = tmp_path / "test.bed"
    bed_file_path.write_text(data)
    return bed_file_path


def test_from_bed_no_filter(bed_file):
    index = GenomicIndex.from_bed(bed_file)

    overlap = index.overlap('chr1', Orientation.fwd, 0, 1000)
    assert overlap.intervals == [Range(0, 1000)]
    assert overlap.annotations == ['gene1']

    overlap = index.overlap('chr1', Orientation.rev, 1000, 1500)
    assert overlap.intervals == [Range(1000, 1500)]
    assert overlap.annotations == ['gene2']

    overlap = index.overlap('chr2', Orientation.fwd, 0, 1000)
    assert overlap.intervals == [Range(0, 500)]
    assert overlap.annotations == ['gene3']

    overlap = index.overlap('chr2', Orientation.rev, 500, 1500)
    assert overlap.intervals == [Range(500, 1000)]
    assert overlap.annotations == ['gene4']


def test_from_bed_with_filter(bed_file):
    index = GenomicIndex.from_bed(bed_file, filterfn=lambda it: it.name in {"gene1", "ABC"})

    overlap = index.overlap('chr1', Orientation.fwd, 940, 10000)
    assert overlap.intervals == [Range(940, 1000)]
    assert overlap.annotations == ['gene1']

    for contig, strand in ('chr1', Orientation.rev), ('chr2', Orientation.fwd), ('chr2', Orientation.rev):
        overlap = index.overlap(contig, strand, 0, 10000)
        assert overlap.intervals == []
        assert overlap.annotations == []



import pytest

from biobit.core import Orientation


def test_orientation_str_repr():
    assert str(Orientation.fwd) == "+"
    assert repr(Orientation.fwd) == "+"
    assert str(Orientation.rev) == "-"
    assert repr(Orientation.rev) == "-"
    assert str(Orientation.dual) == "."
    assert repr(Orientation.dual) == "."


def test_orientation_normalize():
    assert Orientation.normalize("+") == Orientation.fwd
    assert Orientation.normalize(1) == Orientation.fwd
    assert Orientation.normalize(Orientation.fwd) == Orientation.fwd
    assert Orientation.normalize("-") == Orientation.rev
    assert Orientation.normalize(-1) == Orientation.rev
    assert Orientation.normalize(Orientation.rev) == Orientation.rev
    assert Orientation.normalize(".") == Orientation.dual
    assert Orientation.normalize(0) == Orientation.dual
    assert Orientation.normalize(Orientation.dual) == Orientation.dual


def test_orientation_normalize_invalid():
    with pytest.raises(ValueError):
        Orientation.normalize("invalid")



from pathlib import Path

from .repmasker import RepmaskerClassification

GRCh38 = RepmaskerClassification(Path(__file__).parent / "GRCh38.tsv.gz")
GRCm39 = RepmaskerClassification(Path(__file__).parent / "GRCm39.tsv.gz")



import os
from pathlib import Path
from typing import Callable

from attrs import define, field

from biobit.analysis import seqproj
from . import descriptor
from .experiment import Experiment


@define(slots=True, frozen=True, eq=True, order=True, repr=True, hash=True, init=False)
class Project:
    source: seqproj.Project = field()
    experiments: tuple[Experiment, ...] = field(converter=lambda x: tuple(x))

    def __init__(
            self, source: seqproj.Project, resfolder: os.PathLike[str], *,
            bioexp2descriptor: Callable[[seqproj.Experiment], str] = descriptor.from_bioexp,
            ensure_exists: bool = False
    ):
        resfolder = Path(resfolder)
        # if not resfolder.is_dir():
        #     raise ValueError(f"Results folder {resfolder} does not exist")

        experiments = []
        for exp in source.experiments:
            descriptor = bioexp2descriptor(exp)
            experiments.append(Experiment(descriptor, exp, resfolder))

            if ensure_exists:
                experiments[-1].ensure_exists()

        self.__attrs_init__(source, experiments)

    @experiments.validator
    def check_experiments(self, _, value):
        if not value:
            raise ValueError("Project must contain at least one experiment")

    @property
    def ind(self) -> str:
        return self.source.ind

    def ensure_exists(self):
        for exp in self.experiments:
            exp.ensure_exists()



from enum import Enum
from typing import Literal, Union

from .strand import Strand


class Orientation(Enum):
    fwd = "+"
    rev = "-"
    dual = "."

    def __repr__(self):
        return self.value

    def __str__(self):
        return self.value

    @classmethod
    def normalize(cls, orientation: 'OrientationLike') -> 'Orientation':
        match orientation:
            case "+" | 1:
                return Orientation.fwd
            case "-" | -1:
                return Orientation.rev
            case "." | 0:
                return Orientation.dual
            case _ if isinstance(orientation, cls):
                return orientation
            case _ if isinstance(orientation, Strand):
                return cls.from_strand(orientation)
            case _:
                raise ValueError(f"Unknown strand: {orientation}")

    @classmethod
    def from_strand(cls, strand: Strand) -> 'Orientation':
        match strand:
            case Strand.fwd:
                return cls.fwd
            case Strand.rev:
                return cls.rev
            case _:
                raise ValueError(f"Unknown strand: {strand}")


OrientationLike = Union[Strand, Orientation, Literal["+", "-", ".", 1, -1, 0]]
